"""Functions used to read and parse equations written with a pydynamo syntax. There is functions to get the equation type (constant, update, initialisation) and the arguments (variables, constants, indices)."""

import ast
import astunparse
import re

# Functions that are not considered as pydynamo "special functions" 
math_fun = {'sqrt', 'exp', 'log'} # Math functions that we can use
not_parse_fun = math_fun.union(__builtins__) # Merged with all builtins

def change_points_to_underscores(line, variables):
    """Change the points of indices `k` and `j` to underscores.

    Parameters
    ----------
    line : str
        Equation to change.
    variables : iterable((str, str))
        List of (variable name, index name) in the equation.

    Returns
    -------
    str
        Equation changed.
    """ 
    for v, i in variables:
        line = re.sub(f'(?!<\w){v}.{i}(?!\w)', f'{v}_{i}', line)
    return line

def reformat_eq(root, variables):
    """Format a parsed equation to a string equation.

    Parameters
    ----------
    root : ast.Module
        Equation.
    variables : iterable((str, str))
        List of (variable name, index name) in the equation.

    Returns
    -------
    str
        Equation on string format.
    """
    line = astunparse.unparse(root)
    line = change_points_to_underscores(line, variables)
    return line.strip()

def is_eq(root):
    """Determines if a parsed line is an equation or not.
    
    Parameters
    ----------
    root : ast.Module
        Equation.

     Returns
    -------
    bool
        True if `root` is an equation (of type Assign), False otherwise.
    """
    try:
        return isinstance(root.body[0], ast.Assign)
    except:
        return False

def is_cst_eq(root):
    """Determines if a parsed line is an equation for a constant or not.
    
    Parameters
    ----------
    root : ast.Module
        Equation.
    """
    if is_eq(root):
        return isinstance(root.body[0].targets[0], ast.Name)

def is_variable_eq(root):
    """Determines if a parsed line is an equation for a variable or not.
    
    Parameters
    ----------
    root : ast.Module
        Equation.
    """
    if is_eq(root):
        return isinstance(root.body[0].targets[0], ast.Attribute)

def is_update_eq(root):
    """Determines if an equation for a variable is an update equation or not.
    
    Parameters
    ----------
    root : ast.Module
        Equation.
    """
    return is_variable_eq(root) and root.body[0].targets[0].attr == 'k'

def is_init_eq(root):
    """Determines if an equation for a variable is an initialisation equation or not.
    
    Parameters
    ----------
    root : ast.Module
        Equation.
    """
    return is_variable_eq(root) and root.body[0].targets[0].attr == 'i'

def is_eq_of_type(root, eq_type):
    """Determines if an equation is of a certain type ('cst', 'init', 'update').
    
    Parameters
    ----------
    root : ast.Module
        Equation.
    eq_type : str
        Equation type, one of 'cst', 'init', 'update'.
    """
    if is_cst_eq(root):
        return  'cst'  == eq_type
    if is_init_eq(root):
        return 'init'  == eq_type
    if is_update_eq(root):
        return 'update' == eq_type
    return False

def get_pars_update_eq(line):
    """Get parameters of an equation if it's an update equation.
    
    Parameters
    ----------
    line : str
        Equation.
    
    Returns
    -------
    (str, dict)
        Tuple with the name of the updated variable, and a dictionnary containing the arguments and the new equation line.
    """
    root = ast.parse(line)
    assert is_update_eq(root), f'Not an update equation: {line}'
    var = root.body[0].targets[0].value.id
    args = get_var_cst_fun(root.body[0].value)
    return (var,
            {'args': args,
             'line': reformat_eq(root.body[0].value, args['var'])})

def get_pars_init_eq(line):
    """Get parameters of an equation if it's an initialisation equation.
    
    Parameters
    ----------
    line : str
        Equation.
    
    Returns
    -------
    (str, dict)
        Tuple with the name of the initialized variable, and a dictionnary containing the arguments and the new equation line.
    """
    root = ast.parse(line)
    assert is_init_eq(root), f'Not a initialisation equation: {line}'
    var = root.body[0].targets[0].value.id
    args = get_var_cst_fun(root.body[0].value)
    return (var,
            {'args': args,
             'line': reformat_eq(root.body[0].value, args['var'])})

def get_pars_cst_eq(line):
    """Get parameters of an equation if it's an constant equation.
    
    Parameters
    ----------
    line : str
        Equation.
    
    Returns
    -------
    (str, dict)
        Tuple with the name of the constant, and a dictionnary containing the arguments and the new equation line.
    """
    root = ast.parse(line)
    assert is_cst_eq(root), f'Not an constant equation: {line}'
    cst = root.body[0].targets[0].id
    args = get_var_cst_fun(root.body[0].value)
    return (cst,
            {'args': args,
             'line': reformat_eq(root.body[0].value, args['var'])})

def get_var_cst_fun(root):
    """Get every constants, variables and special functions contained in the equation.
    
    Parameters
    ----------
    root : ast.Module
        Equation.
    
    Returns
    -------
    dict(str: set)
        A dictionnary which contains the set of constants, variables and functions names.
    {'cst': set(str), 'var': set((str, str)), 'fun': dict(str: {'args': list(str), 'type': str, 'fun': str})}
    """
    list_params = list()
    all_names = set()
    all_vars = set()
    all_nps_funs = set()
    all_fun = dict()
    for node in ast.walk(root):
        if isinstance(node, ast.Name):
            all_names.add(node.id)
        elif isinstance(node, ast.Attribute):
            all_vars.add((node.value.id, node.attr))
        elif isinstance(node, ast.Call):
            assert isinstance(node.func, ast.Name), f"Function call can only be a name: {astunparse.unparse(node.func)}"
            if node.func.id not in not_parse_fun:
                arg_fun = {'args': [astunparse.unparse(arg)
                                              for arg in node.args],
                           'type': node.func.id,
                           'fun': node.func.id}
                all_fun[node.func.id] = arg_fun
            else:
                all_nps_funs.add(node.func.id)
    return {'cst': {c for c in all_names.difference(v for v, i in all_vars).difference(all_fun).difference(all_nps_funs)
                    if c not in math_fun},
            'var': all_vars,
            'fun': all_fun}

def get_pars_eq(root, eq_type):
    """Get parameters of an equation.
    
    Parameters
    ----------
    root : ast.Module
        Equation.
    eq_type : str
        Equation type, one of 'cst', 'init', 'update'.
    Returns
    -------
    (str, dict) 
        Variable or constant name, and parameters.
    """
    if 'cst'  == eq_type:
        return get_pars_cst_eq(root)
    elif 'update'  == eq_type:
        return get_pars_update_eq(root)
    elif 'init'  == eq_type:
        return get_pars_init_eq(root)
    return (None,
            {'var': set(),
             'fun': set()})



# WORLD3 2003 updated equations, in pydynamo syntax.
# Population sector
pop.k = p1.k + p2.k + p3.k + p4.k
p1.k = p1.j + dt*(b.j - d1.j - mat1.j)
p1.i = p1i
p1i = 65e7
d1.k = p1.k * m1.k
m1.k = tabhl(m1t, le.k, 20, 80, 10)
m1t = [0.0567, 0.0366, 0.0243, 0.0155, 0.0082, 0.0023, 0.001]
mat1.k = p1.k*(1 - m1.k)/15
p2.k = p2.j + dt*(mat1.j - d2.j - mat2.j)
p2.i = p2i
p2i = 70e7
d2.k = p2.k * m2.k
m2.k = tabhl(m2t, le.k, 20, 80, 10)
m2t = [0.0266, 0.0171, 0.0110, 0.0065, 0.0040, 0.0016, 0.0008]
mat2.k = p2.k*(1 - m2.k)/30
p3.k = p3.j + dt*(mat2.j-d3.j-mat3.j)
p3.i = p3i
p3i = 19e7
d3.k = p3.k * m3.k
m3.k = tabhl(m3t, le.k, 20, 80, 10)
m3t = [0.0562, 0.0373, 0.0252, 0.0171, 0.0118, 0.0083, 0.006]
mat3.k = p3.k*(1 - m3.k)/20
p4.k = p4.j + dt*(mat3.j - d4.j)
p4.i = p4i
p4i = 6e7
d4.k = p4.k * m4.k
m4.k = tabhl(m4t, le.k, 20, 80, 10)
m4t = [0.13, 0.11, 0.09, 0.07, 0.06, 0.05, 0.04]

# Death rate subsector
d.k = d1.j + d2.j + d3.j + d4.j
cdr.k = 1000*d.k/pop.k # death rate
le.k = len*lmf.k*lmhs.k*lmp.k*lmc.k
len = 28
lmf.k = tabhl(lmft, fpc.k/sfpc, 0, 5, 1)
lmft = [0, 1, 1.43, 1.5, 1.5, 1.5]
hsapc.k = tabhl(hsapct, sopc.k, 0, 2000, 250)
hsapct = [0, 20, 50, 95, 140, 175, 200, 220, 230]
ehspc.k = smooth(hsapc.j, hsid)
hsid = 20
lmhs.k = clip(lmhs1.k, lmhs2.k, 1940, time.k)
lmhs1.k = tabhl(lmhs1t, ehspc.k, 0, 100, 20)
lmhs1t = [1, 1.1, 1.4, 1.6, 1.7, 1.8]
lmhs2.k = tabhl(lmhs2t, ehspc.k, 0, 100, 20)
lmhs2t = [1, 1.5, 1.9, 2, 2, 2]
fpu.k = tabhl(fput, pop.k, 0, 16e9, 2e9)
fput = [0, 0.2, 0.4, 0.5, 0.58, 0.65, 0.72, 0.78, 0.80]
cmi.k = tabhl(cmit, iopc.k, 0, 1600, 200)
cmit = [0.5, 0.05, -0.1, -0.08, -0.02, 0.05, 0.1, 0.15, 0.2]
lmc.k = 1 - (cmi.k*fpu.k)
lmp.k = tabhl(lmpt, ppolx.k, 0, 100, 10)
lmpt = [1, 0.99, 0.97, 0.95, 0.90, 0.85, 0.75, 0.65, 0.55, 0.40, 0.20]

# Birth rate subsector
b.k = clip((tf.k*p2.k*0.5/rlt), d.k, pet, time.k)
cbr.k = 1000*b.k/pop.k # birth rate 
rlt = 30
pet = 4000
cbr.k = 1000*b.j/pop.k
tf.k = min(mtf.k, (mtf.k*(1-fce.k) + dtf.k*fce.k))
mtf.k = mtfn * fm.k
mtfn = 12
fm.k = tabhl(fmt, le.k, 0, 80, 10)
fmt = [0, 0.2, 0.4, 0.6, 0.7, 0.75, 0.79, 0.84, 0.87]
dtf.k = dcfs.k*cmple.k
cmple.k = tabhl(cmplet, ple.k, 0, 80, 10)
cmplet = [3, 2.1, 1.6, 1.4, 1.3, 1.2, 1.1, 1.05, 1]
ple.k = dlinf3(le.k, lpd)
lpd = 20
dcfs.k = clip(dcfsn*frsn.k*sfsn.k, 2.0, zpgt, time.k)
zpgt = 4000
dcfsn = 3.8
sfsn.k = tabhl(sfsnt, diopc.k, 0, 800, 200)
sfsnt =  [1.25, 0.94, 0.715, 0.59, 0.5] # Changed from Vensim. Before it was  [1.25, 1, 0.9, 0.8, 0.75]
diopc.k = dlinf3(iopc.k, sad) 
sad = 20
frsn.k = tabhl(frsnt, fie.k, -0.2, 0.2, 0.1)
frsnt = [0.5, 0.6, 0.7, 0.85, 1]
# frsn.i = 0.82 # Removed from Vensim 
fie.k = (iopc.k - aiopc.k)/aiopc.k
aiopc.k = smooth(iopc.j, ieat)
ieat = 3
nfc.k = (mtf.k / dtf.k) - 1
fce.k = clip(tabhl(fcet, fcfpc.k, 0, 3, 0.5), 1.0, fcest, time.k)
fcest = 4000
fcet = [0.75, 0.85, 0.9, 0.95, 0.98, 0.99, 1]
fcfpc.k = dlinf3(fcapc.k, hsid)
fcapc.k = fsafc.k*sopc.k
fsafc.k = tabhl(fsafct, nfc.k, 0, 10, 2)
fsafct = [0, 0.005, 0.015, 0.025, 0.03, 0.035]

# Capital sector
## Industrial subsector
iopc.k = io.k/pop.k
io.k = ic.k*(1-fcaor.k)*cuf.k/icor.k
icor.k = clip(icor1, icor2.k, pyear, time.k)
icor1 = 3
icor2.k = icormrrct.k*icormlyt.k*icormpt.k
icormrrct.k = tabhl(icormrrctt, nruf.j,0, 1, 0.1) # industrial capital output ratio multiplier from resource conservation technology
icormrrctt = [3.75, 3.6, 3.47, 3.36, 3.25, 3.16, 3.1, 3.06, 3.02, 3.01, 3] # industrial capital output ratio multiplier from resource table
icormlyt.k = tabhl(icormlytt, lymt.j, 1, 2, 0.2) # industrial capital output ratio multiplier from land yield technology
# Note: Weird stuff in Vensim file (1,0.8)-(2,2)],
icormlytt = [1, 1.05, 1.12, 1.25, 1.35, 1.5] # industrial capital output ratio multiplier table
icormpt.k = tabhl(icormptt, ppgf.j, 0, 1, 0.1) # industrial capital output ratio multiplier from pollution technology
icormptt = [1.25, 1.2, 1.15, 1.11, 1.08, 1.05, 1.03, 1.02, 1.01, 1, 1] # industrial capital output ratio multiplier from pollution table
ic.k = ic.j + dt*(icir.j-icdr.j)
ic.i = ici
ici = 2.1e11
icdr.k = ic.k/alic.k
alic.k = clip(alic1, alic2, pyear, time.k)
alic1 = 14
alic2 = 14
icir.k = io.k*fioai.k
fioai.k = 1- fioaa.k - fioas.k - fioac.k
fioac.k = clip(fioacc.k, fioacv.k, iet, time.k)
iet = 4000
fioacc.k = clip(fioac1, fioac2, pyear, time.k)
fioac1 = 0.43
fioac2 = 0.43
fioacv.k = tabhl(fioacvt, iopc.k/iopcd, 0, 2, 0.2)
fioacvt = [0.3, 0.32, 0.34, 0.36, 0.38, 0.43, 0.73, 0.77, 0.81, 0.82, 0.83]
iopcd = 400

## Service subsector
isopc.k = clip(isopc1.k, isopc2.k, pyear, time.k)
isopc1.k = tabhl(isopc1t, iopc.k, 0, 1600, 200)
isopc1t = [40, 300, 640, 1000, 1220, 1450, 1650, 1800, 2000]
isopc2.k = tabhl(isopc2t, iopc.k, 0, 1600, 200)
isopc2t = [40, 300, 640, 1000, 1220, 1450, 1650, 1800, 2000]
fioas.k = clip(fioas1.k, fioas2.k, pyear, time.k)
fioas1.k = tabhl(fioas1t, sopc.k/isopc.k, 0, 2, 0.5)
fioas1t = [0.3, 0.2, 0.1, 0.05, 0]
fioas2.k = tabhl(fioas2t, sopc.k/isopc.k, 0, 2, 0.5)
fioas2t = [0.3, 0.2, 0.1, 0.05, 0]
scir.k = io.k*fioas.k
sc.k = sc.j + dt*(scir.j-scdr.j)
sc.i = sci
sci = 1.44e11
scdr.k = sc.k/alsc.k
alsc.k = clip(alsc1, alsc2, pyear, time.k)
alsc1 = 20
alsc2 = 20
so.k = (sc.k*cuf.k)/scor.k
sopc.k = so.k/pop.k
scor.k = clip(scor1, scor2, pyear, time.k)
scor1 = 1
scor2 = 1

## Job subsector
j.k = pjis.k + pjas.k +pjss.k
pjis.k = ic.k*jpicu.k
jpicu.k = tabhl(jpicut, iopc.k, 50, 800, 150)*1e-3
jpicut = [0.37, 0.18, 0.12, 0.09, 0.07, 0.06]
pjss.k = sc.k*jpscu.k
jpscu.k = tabhl(jpscut, sopc.k, 50, 800, 150)*1e-3
jpscut = [1.1, 0.6, 0.35, 0.2, 0.15, 0.15]
pjas.k = jph.k*al.k
jph.k = tabhl(jpht, aiph.k, 2, 30, 4)
jpht = [2, 0.5, 0.4, 0.3, 0.27, 0.24, 0.2, 0.2]
lf.k = (p2.k + p3.k)*lfpf
lfpf = 0.75
luf.k = j.k/lf.k
lufdi.k = clip(1, luf.j, initial_time, time.k) # helper for luf
lufd.k = smooth(lufdi.k, lufdt) # Added from  Vensim
lufdt = 2
cuf.k = tabhl(cuft, lufd.k, 1, 11, 2)
cuf.i = 1
cuft = [1, 0.9, 0.7, 0.3, 0.1, 0.1]

# Agricultural sector
## Loop1: food from investment in land development
lfc.k = al.k/palt
palt = 3.2e9
al.k = al.j + dt*(ldr.j - ler.j - lrui.j)
al.i = ali
ali = 0.9e9
pal.k = pal.j + dt*(-ldr.j)
pal.i = pali
pali = 2.3e9
f.k = ly.k*al.k*lfh*(1-pl)
lfh = 0.7
pl = 0.1
fpc.k = f.k/pop.k
ifpc.k = clip(ifpc1.k, ifpc2.k, pyear, time.k)
ifpc1.k = tabhl(ifpc1t, iopc.k, 0, 1600, 200)
ifpc1t = [230, 480, 690, 850, 970, 1070, 1150, 1210, 1250]
ifpc2.k = tabhl(ifpc2t, iopc.k, 0, 1600, 200)
ifpc2t = [230, 480, 690, 850, 970, 1070, 1150, 1210, 1250]
tai.k = io.k*fioaa.k
fioaa.k = clip(fioaa1.k, fioaa2.k, pyear, time.k)
fioaa1.k = tabhl(fioaa1t, fpc.k/ifpc.k, 0, 2.5, 0.5)
fioaa1t = [0.4, 0.2, 0.1, 0.025, 0, 0]
fioaa2.k = tabhl(fioaa2t, fpc.k/ifpc.k, 0, 2.5, 0.5)
fioaa2t = [0.4, 0.2, 0.1, 0.025, 0, 0]
ldr.k = tai.k*fiald.k/dcph.k
dcph.k = tabhl(dcpht, pal.k/palt, 0, 1, 0.1)
dcpht = [1e5, 7400, 5200, 3500, 2400, 1500, 750, 300, 150, 75, 50]

## Loop2: food from investment in agricultural inputs
cai.k = tai.k * (1 - fiald.k)
ai.k = smooth(cai.j, alai.k)
ai.i = 5e9
alai.k = clip(alai1, alai2, pyear, time.k)
alai1 = 2
alai2 = 2
aiph.k = ai.k*(1 - falm.k)/al.k
lymc.k = tabhl(lymct, aiph.k, 0, 1000, 40)
lymct = [1, 3, 4.5, 5.0, 5.3, 5.6, 5.9, 6.1, 6.35, 6.6, 6.9, 7.2, 7.4, 7.6, 7.8, 8, 8.2, 8.4, 8.6, 8.8, 9, 9.2, 9.4, 9.6, 9.8, 10]
ly.k = lymt.k*lfert.k*lymc.k*lymap.k
lymt.k = clip(lyf1, lyf2.k, pyear, time.k) # land yield multiplier from technology
lyf1 = 1
lyf2.k = smooth(lyt.k, tdd)
lyt.k = lyt.j + dt*(lytcr.k) # land yield technology
lyt.i = 1
lytcr.k = clip(0, lyt.j*lytcrm.j, pyear, time.k)# land yield technology change rate
lytcrm.k = tabhl(lytcrmt, drf - fr.k, 0, 1, 1)# land yield technology change rate multiplier
lytcrmt = [0, 0] # land yield technology change rate multiplier table
drf = 2 # desired food ratio
lymap.k = clip(lymap1.k, lymap2.k, appyear, time.k)
appyear = 4000 # air pollution policy implementation time
lymap1.k = tabhl(lymap1t, io.k/io70, 0, 30, 10)
lymap1t = [1, 1, 0.7, 0.4]
lymap2.k = tabhl(lymap2t, io.k/io70, 0, 30, 10)
lymap2t = [1, 1, 0.98, 0.95] # Seen in wrld3+03.mdl
io70 = 7.9e11

## Loop 1 & 2: the investment allocation decision
fiald.k = tabhl(fialdt, mpld.k/mpai.k, 0, 2, 0.25)
fialdt = [0, 0.05, 0.15, 0.30, 0.50, 0.70, 0.85, 0.95, 1]
mpld.k = ly.k/(dcph.k*sd)
sd = 0.07
mpai.k = alai.k*ly.k*mlymc.k/lymc.k
mlymc.k = tabhl(mlymct, aiph.k, 0, 600, 40)
mlymct = [0.075, 0.03, 0.015, 0.011, 0.009, 0.008, 0.007, 0.006, 0.005, 0.005, 0.005, 0.005, 0.005, 0.005, 0.005, 0.005]

## Loop 3: land erosion and urban-industrial use
all.k = alln*llmy.k
alln = 1000
llmy.k = clip(llmy1.k, 0.95**((time.k - llmytm)/oy)*llmy1.k + (1 - 0.95**((time.k - llmytm)/oy))*llmy2.k, llmytm, time.k) 
llmy1.k = tabhl(llmy1t, ly.k/ilf, 0, 9, 1)
llmy1t = [1.2, 1, 0.63, 0.36, 0.16, 0.055, 0.04, 0.025, 0.015, 0.01]
llmy2.k = tabhl(llmy2t, ly.k/ilf, 0, 9, 1)
llmy2t = [1.2, 1, 0.63, 0.36, 0.29, 0.26, 0.24, 0.22, 0.21, 0.2] # Note: added from Vensim
ler.k = al.k/all.k
uilpc.k = tabhl(uilpct, iopc.k, 0, 1600, 200)
uilpct = [0.005, 0.008, 0.015, 0.025, 0.04, 0.055, 0.07, 0.08, 0.09]
uilr.k = uilpc.k*pop.k
lrui.k = max(0, (uilr.k - uil.k)/uildt)
uildt = 10
uil.k = uil.j + dt*(lrui.j)
uil.i = uili
uili = 8.2e6

## Loop 4: land fertility degradation
lfert.k = lfert.j + dt*(lfr.j-lfd.j)
lfert.i = lferti
lferti = 600
lfdr.k = tabhl(lfdrt, ppolx.k, 0, 30, 10)
lfdrt = [0, 0.1, 0.3, 0.5]
lfd.k = lfert.k*lfdr.k

## Loop 5: land fertility regeneration
lfr.k = (ilf-lfert.k)/lfrt.k
ilf = 600
lfrt.k = tabhl(lfrtt, falm.k, 0, 0.10, 0.02)
lfrtt = [20.0, 13.0, 8.0, 4.0, 2.0, 2.0]

## Loop 6: Discontinuing land maintinance
falm.k = tabhl(falmt, pfr.k, 0, 4, 1)
falmt = [0.0, 0.04, 0.07, 0.09, 0.1]
fr.k = fpc.k/sfpc
sfpc = 230
pfr.k = smooth(fr.j, fspd)
pfr.i = 1
fspd = 2

# Nonrenewable resource sector
nr.k = nr.j + dt*(-nrur.j)
nr.i = nri
nri = 1e12
nrur.k = pop.k * pcrum.k * nruf.k
nruf.k = clip(nruf1, nruf2.k, pyear, time.k)
nruf1 = 1
nruf2.k = smooth(rct.k, tdd)
rct.k = rct.j - dt*(rctcr.k)# resource conservation technology
rct.i = 1
rctcr.k = clip(0, rct.j*rtcm.j, pyear, time.k) # resource technology change rate
rtcm.k = tabhl(rtcmt,1-nrur.j/drur, -1, 0, 1) # resource technology change rate multiplier
drur = 4.8e09 # desired resource use rate
rtcmt = [0, 0]# resource technology change mult table
pcrum.k = tabhl(pcrumt, iopc.k, 0, 1600, 200)
pcrumt = [0, 0.85, 2.6, 3.4, 3.8, 4.1, 4.4, 4.7, 5]
nrfr.k = nr.k/nri
fcaor.k = clip(fcaor1.k, fcaor2.k, fcaortm, time.k)
fcaor1.k = tabhl(fcaor1t, nrfr.k, 0, 1, 0.1)
fcaor1t = [1.0, 0.9, 0.7, 0.5, 0.2, 0.1, 0.05, 0.05, 0.05, 0.05, 0.05]
fcaor2.k = tabhl(fcaor2t, nrfr.k, 0, 1, 0.1)
fcaor2t = [1.0, 0.2, 0.1, 0.05, 0.05, 0.05, 0.05, 0.05, 0.05, 0.05, 0.05]


# Persistent pollution sector
ppgr.k = (ppgio.k + ppgao.k)*ppgf.k
ppgf.k = clip(ppgf1, ppgf2.k, pyear, time.k)
ppgf1 = 1
ppgf2.k = smooth(ppt.k, tdd)
tdd = 20 # technology development delay
ppt.k = ppt.j + dt*pptcr.k # persistent pollution technology
ppti = 1 # Initial persistent pollution technology
ppt.i = ppti
pptcr.k = clip(0, ppt.j*pptcm.j, pyear, time.k) # persistent pollution technology change rate
pptcm.k = tabhl(pptcmt, 1 - ppolx.k/dppolx, -1, 0, 1)# persistent pollution technology change multiplier
pptcmt = [0, 0] # persistent pollution technology change mult table
dppolx = 1.2 # desired persistent pollution index
# Note: industrial capital output ratio multiplier from persistent pollution technology is not implemented in Vensim while on the schemas
ppgio.k = pcrum.k * pop.k * frpm*imef*imti
frpm = 0.02
imef = 0.1
imti = 10
ppgao.k = aiph.k*al.k*fipm*amti
fipm = 0.001
amti = 1
ppapr.k = delay3(ppgr.j, pptd) # Delay3i ??
pptd = 20
ppol.k = ppol.j + dt*(ppapr.j-ppasr.j)
ppol.i = ppoli
ppoli = 2.5e7
ppolx.k = ppol.k/ppol70
ppol70 = 1.36e8
ppasr.k = ppol.k/(ahl.k*1.4)
ahlm.k = tabhl(ahlmt, ppolx.k, 1, 1001, 250)
ahlmt = [1.0, 11.0, 21.0, 31.0, 41.0]
ahl.k = ahl70*ahlm.k
ahl70 = 1.5

# World3 03 supplementary equations sector
cio.k = io.k*fioacc.k # consumed industrial output
ciopc.k = cio.k/pop.k # consumed industrial output per capita
fao.k = pof * f.k/(pof*f.k + so.k + io.k) # fraction of output in agriculture
fos.k = so.k/(pof*f.k + so.k + io.k) # fraction of output in services
foi.k = io.k/(pof*f.k + so.k + io.k) # fraction of output in industry
llmytm = 4000 # land life policy implementation time
pof = 0.22 # price of food
fcaortm = 4000# fraction of industrial capital allocated to obtaining resources switch time
plinid.k = ppgio.k*ppgf.k/io.k  # persistent pollution intensity industry
resint.k = nrur.k/io.k # resource use intensity
thousand = 1000 # THOUSAND

# World3 03 indicators sector
ugdp = 1 # GDP pc unit
uai = 1 # unit agricultural input
up = 1 # unit population
ablgha.k = ppgr.k*hpup/hpgha # "Absorption Land (GHA)"
ei.k = tabhl(eit, gdppc.k/ugdp, 0, 7000, 1000) # Education Index
eit = [0, 0.81, 0.88, 0.92, 0.95, 0.98, 0.99, 1] # Education Index LOOKUP
gdpi.k = log(gdppc.k/rlgdp, 10)/log(rhgdp/rlgdp, 10) # GDP index
gdppc.k = tabhl(gdppct, iopc.k/ugdp, 0, 1000, 200) # GDP per capita
gdppct = [120, 600, 1200, 1800, 2500, 3200] # GDP per capita LOOKUP
hpgha = 1e9 # ha per gha
hpup = 4 # ha per unit of pollution
hef.k = (algha.k + ulgha.k + ablgha.k)/tl # Human Ecological Footprint
hwi.k = (lei.k + ei.k + gdpi.k)/3 # Human Welfare Index
lei.k = tabhl(leit, le.k/oy, 25, 85, 10) # Life Expectancy Index
leit = [0, 0.16, 0.33, 0.5, 0.67, 0.84, 1]#  Life Expectancy Index LOOKUP
oy = 1 # one year
rhgdp = 9508 # Ref Hi GDP
rlgdp = 24 # Ref Lo GDP
tl = 1.91 # Total land
ulgha.k = uil.k/hpgha # "Urban Land (GHA)"
algha.k = al.k/hpgha # "Arable Land in Gigahectares (GHA)"

# Control card for simulation
pyear = 1975
initial_time = 1900

# Take care to keep the last line :)



"""System class which creates a System from a Pysd model.
"""

import networkx as nx
import numpy as np
import re
from .system import System

class PsdSystem(System):
    """System which initialises with a Pysd model.
    """
    def __init__(self, model):
        """Intialise from a Pysd model. Comments are retrieved from code.

        Parameters
        ----------
        mode : PySD.model
            Model from which the System is created.

        """
        self.model = model
        self.caracs = {}
        self.comments = {}
        
        for n in self.model.components._dependencies:
            try :
                doc = getattr(self.model.components, n).__doc__
                car = {}
                coms = []
                for l in doc.split('\n'):
                    if ':' in l:
                        key, val = l.split(':')
                        car[key.strip()] = val.strip()
                    else:
                        coms.append(l)
                car['Comment'] = re.sub('  *', ' ', ' '.join(coms))
                self.caracs[n] = car
                
                try:
                    self.comments[n] = f"{car['Real Name']} [{car['Units']}]: {car['Comment']}"
                except Exception as e:
                    if any(i in n for i in {'smooth', 'integ', 'delay'}):
                        self.comments[n] = ''
                    else:
                        print(f'Counlnt find comment for var {n}', e)
                        raise e
            except Exception as e:
                if 'active' not in n:
                    print(f'Problem for car of {n}')
                    raise e
            
        self.df_run = None
        
    def get_influence_graph(self):
        """Get the graph of influences: an arrow from A to B if B needs A (at initialisation or updating step) to be computed.

    Returns
    -------
    networkx.DiGraph
        Graph of influences.
    """
        G = nx.Graph()
        for var, deps in self.model.components._dependencies.items():
            for dep, i in deps.items():
                if i is not None:
                    G.add_edge(dep, var)
        return G

    def get_tabhl_args(self, name):
        """
        Get indications about a tabhl function.
        
        Parameters
        ----------
        name: str
            Name of the variable using a tabhl function.

        Returns
        -------
        np.array, np.array, str, str, str:
            x, f(x), x label, y label, title
        """
        
        if self.caracs[name]['Type'] == 'lookup':
            try:
                x, y = np.array(eval(self.caracs[name]['Original Eqn'])).T
                ylabel = self.caracs[name]['Units']
                return x, y, ylabel, '', ''
            except Exception as e:
                print('no such !')
                raise e
        return None, None, '', '', ''
    
    def get_var_names(self):
        """Return every variables names.
        
        Returns
        -------
        list(str)
            List of variables names.
        """
        return list(self.model.components._dependencies.keys())

    def get_time(self):
        """Returns the time array.

        Returns
        -------
        np.array(float):
            Array of system time.
        """
        return np.arange(self.model.time.initial_time(), self.model.time.final_time()+self.model.time.time_step()/2, self.model.time.time_step())

    def get_var(self, name):
        """Get the variable array.

        Parameters
        ----------
        name: str
            Name of variable.

        Returns
        -------
        np.array(float):
            Array of values of the variable for the last run.
        """
        if self.df_run is None:
            return self.get_time()*0
        else:
            return self.df_run[name]

    def run(self):
        """Run PySD model.
        """
        self.df_run = self.model.run()

    def equation(self, name):
        """Returns the reformatted equation of a variable or constant.

        Parameters
        ----------
        node : str
            Name of the variable or constant to show the equation.
        """
        try:
            return self.caracs[name]['Original Eqn']
        except:
            return ""



"""Functions implementing a delayed smooth effect in pydynamo.
"""
import numpy as np

class Smooth:
    """
    Delay information function of the 1st order for smoothing.
    Returns a class that is callable as a function (see Call parameters) at a given step k.

    Parameters
    ----------
    dt : float
        time step.

    Call parameters
    ---------------
    delay_time : float
        delay parameter. Higher delay increases smoothing.
    val : float
        current value to smooth
    k : int
        current loop index.
    """
    def __init__(self, dt):
        self.state = None
        self.k = 0
        self.dt = dt

    def initialise(self, init_value):
        self.state = init_value

    def ddt(self, value, delay_time):
        self.state = self.state + (value - self.state)/delay_time*self.dt
    
    def __call__(self, val_k, delay_time, k):
        if self.state is None:
            if k == 0 or k == 1:
                self.initialise(val_k)
            else:
                raise Exception('Bad call of smooth')
            
        if k == self.k + 1:
            res = self.state
            self.ddt(val_k, delay_time)
            self.k = self.k + 1
            return res
        
        if k != self.k:
            raise Exception("Bad call of smooth function")
        
        return self.state

class Dlinf3:
    """
    Delay information function of the 3st order for smoothing. 
    Returns a class that is callable as a function (see Call parameters) at a given step k.

    Parameters
    ----------
    dt : float
        time step.

    Call parameters
    ---------------
    delay_time : float
        delay parameter. Higher delay increases smoothing.
    val : float
        current value to smooth
    k : int
        current loop index.
    """
    def __init__(self, dt):
        self.state = None
        self.k = 0
        self.dt = dt

    def initialise(self, init_value):
        self.state = np.ones(3)*init_value

    def ddt(self, value, delay_time):
        targets = np.roll(self.state, 1, axis=0)
        targets[0] = value
        
        self.state = self.state + (targets - self.state)*3/delay_time*self.dt
    
    def __call__(self, val_k, delay_time, k):
        if self.state is None:
            self.initialise(val_k)
                
        if k == self.k + 1:
            self.ddt(val_k, delay_time)
            self.k = self.k + 1
            
        if k != self.k:
            raise Exception("Bad call of dlinf3 function")
        
        return self.state[-1]
    
class Delay3:
    """
    Delay function of the 3rd order. Returns a class that is callable as a
    function (see Call parameters) at a given step k.
    
    Parameters
    ----------
    dt : float
        time step.

    Call parameters
    ---------------
    delay_time : float
        delay parameter. Higher delay increases smoothing.
    val_k : float
        current value to delay
    k : int
        current loop index.
    """

    def __init__(self, dt):
        self.state = None
        self.k = 0
        self.dt = dt

    def initialise(self, init_value, delay_time):
        self.state = np.ones(3)*init_value*delay_time

    def ddt(self, value, delay_time):
        outf = self.state/delay_time
        inf = np.roll(outf, 1, axis=0)
        inf[0] = value
        self.state = self.state + (inf - outf)*3*self.dt
    
    def __call__(self, val_k, delay_time, k):
        if self.state is None:
            self.initialise(val_k, delay_time)
                
        if k == self.k + 1:
            self.ddt(val_k, delay_time)
            self.k = self.k + 1
            
        if k != self.k:
            raise Exception("Bad call of delay function")
        
        return self.state[-1]/delay_time

Dlinf3 = Delay3



"""Infomations about sectors of each variable.
"""

import re

def get_sectors(fname):
    """From the world3 pydynamo code get all sectors for each variable.
    
    Parameters
    ----------
    fname : str
        World3 pydynamo code file.
    """
    sectors = {}
    subsectors = {}

    with open(fname, 'r') as f:
        for l in f.readlines():
            if l.startswith('#'):
                if ' sector' in l:
                    sector = l[1:].strip()
                    subsector = ''
                elif ' subsector' in l:
                    subsector = l[1:].strip()

            else:
                try:
                    var = re.findall('^(\w*)[\. \=]', l)[0]
                    sectors[var] = sector
                    subsectors[var] = subsector

                except Exception as e:
                    assert l.strip() == '', 'error: ' + l + '\n' + repr(e)
    return sectors, subsectors

from collections import defaultdict
sector_color = defaultdict(lambda: '#888888', **{
    'Persistent pollution sector': '#19D9FF',
    'Population sector': '#CF4125',
    'Capital sector': '#0073E6',
    'Agricultural sector': '#66CC00',
    'Nonrenewable resource sector': '#8080FF',
})


# # Dict(variable: corresponding sector)
# var_sector, var_subsector = get_sectors('../limits_to_growth_pydynamo_code.py')

# # Dict(sector: [corresponding variables]
# sector_vars = {sector: [] for sector in var_sector.values()}
# for var in var_sector:
#     sector_vars[var_sector[var]].append(var)

# # Dict(variable: corresponding color)
# var_color = {v: sector_color[sector] for v, sector in var_sector.items()}

# Note: variables but also constants and tables are included
# import json

# for dico, filename in ((var_sector, 'var_sector.json'),
#                        (sector_vars, 'sector_vars.json'),
#                        (var_color, 'var_color.json')):
#     with open(filename, 'w+') as f:
#         json.dump(dico, f)



"""DEPRECIATED. Some functions to read definitions from a file in markdown format, usefull for translation.
"""
def read_translated_defs(filename):
    """Reads definitions in markdown format (see begin_sentence)
    and returns the definitions dictionnary.
    
    Parameters
    ----------
    filename : str
        File in which are stored the definitions in Markdown format.
    """
    defs = {}
    with open(filename, 'r') as f:
        var_name = ''
        for l in f.readlines():
            if l.startswith('##'):
                var_name = l[2:].strip()
                defs[var_name] = []
            elif l.startswith('- '):
                defs[var_name].append(l[2:].strip())
    return defs

begin_sentence = """# World3 definitions
Definitions in french and english, but differently formatted,
in order to write transalations by hand more easily.
Go to file get_definitions.py to convert this file to jso.

Syntax:
Variable name should begin by a '## ' and definitions in english and then french are at the two next lines begining by a '- '.
"""

def write_translated_defs(filename, defs):
    """Write definitions that are in the dictionnary defs
    in markdown format in the filename.

    Parameters
    ----------
    filename : str
        File in which to write the definitions in markdown format.
    defs : dict
        Definitions.
    """
    
    with open(filename, 'w+') as f:
       f.write(begin_sentence)
       for k in defs:
           f.write(f'## {k}\n')
           f.write(f'- {defs[k][0]}\n')
           f.write(f'- {defs[k][1]}\n')

        
# import json

# From markdown to definitions dict
# translated_defs = read_translated_defs('translated_defs.md')

# From definitions dict to markdown
# with open('translated_defs.json', 'w+') as f:
    # json.dump(translated_defs, f)

# # From json to dict    
# with open('translated_defs.json', 'r') as f:
#     translated_defs = json.load(f)
    



"""Functions to parse an entire pydynamo code and generate a System object. Also defines every political changes.
"""
import ast
import inspect

from .parse_equations import *
from .parse_dynamo_functions import get_dynamo_fun_params
from .specials import step, clip

def comment_from_equation(line):
    """Retrieve comment by removing '#' and additional spaces.
    """
    if '#' in line:
        return line.split('#', 1)[1].strip()
    return ''

def get_system_dicts(lines):
    """

    Parameters
    ----------
    lines : iterable(str)
        List of every equations.
    
    Returns
    -------

    (dict, dict, dict)
        Nodes (variable, constants and functions), equations (constant, update and initialisation) and comments parsed in the equations list.
    """
    
    all_eqs = {name: dict() for name in ['cst', 'update', 'init']}
    all_nodes = {name: set() for name in ['cst', 'var', 'fun']}
    comments = {}

    # For each equation, try to detect its type and retrive informations.
    for l in lines:
        root = ast.parse(l)
        type_is_identified = False
        
        for eq_type in all_eqs.keys():
            if is_eq_of_type(root, eq_type):
                type_is_identified = True

                # Get assigned node and equation arguments
                try:
                    node, args = get_pars_eq(root, eq_type)
                    args['raw_line'] = l.split('#')[0].strip()
                except:
                    raise Exception("Error while parsing line:\n"+l)
                
                # Add comment
                com = comment_from_equation(l)
                if com != '' or node not in comments:
                    comments[node] = com

                # Add equation
                all_eqs[eq_type][node] = args

                # Handle special functions
                new_line, fun_args = get_dynamo_fun_params(root, node)
                if new_line:
                    all_eqs[eq_type][node]['line'] = reformat_eq(new_line, args['args']['var'])
                if fun_args:
                    all_eqs[eq_type][node]['args']['fun'][fun_args['type']] = fun_args

                # Add assigned node
                if eq_type == 'cst':
                    all_nodes['cst'].add(node)
                else:
                    all_nodes['var'].add(node)

                # Add needed nodes
                for arg_node in args['args']['fun']:
                    if arg_node not in __builtins__:
                        all_nodes['fun'].add(arg_node)

                            
        if root.body and not type_is_identified:
            raise(SyntaxError(f"Invalid equation:\n {l}"))

    return all_nodes, all_eqs, comments


def list_from_file(filename, s=None, prepare=True):
    """Get a list of equatinos from a file with pydynamo equations.
    Parameters
    ----------
    filename : str
        Files in which every pydynamo equations are written.
    """
    check_file(filename)
    with open(filename, 'r') as f:
        return f.readlines()

def list_from_function(fun, s=None, prepare=True):
    """Get a list of equations from a function which lists pydynamo equations.

    Parameters
    ----------
    fun : function
        Function in which every pydynamo equations are written.

    Examples 
    --------
    Just write pydynamo equations inside a function, and create a System with it:
    >>> def custom_equations():
    >>>     pop.i = 100
    >>>     pop.k = pop.j /2 # Population
    >>> list_of_equations = list_from_function(custom_equations)
    """
    return [l.strip() for l in inspect.getsource(fun).split('\n')[1:]]

def check_file(filename):
    """Check if every line in the file can be parsed by the ast module. If not, raise an error.
    Parameters
    ----------
    filename : str
        Files in which every pydynamo equations are written.
    """
    with open(filename, 'r') as f:
        for i, l in enumerate(f.readlines()):
            try:
                ast.parse(l)
            except SyntaxError as e:
                e.filename = filename
                e.lineno = i + 1
                raise



"""
Data used to build world3 model.

Pydynamo code, colors, definitions
"""

from .get_sectors import get_sectors, sector_color

import os
import json

code_file = os.path.join(os.path.dirname(__file__),'code_pydynamo_w3.py')

w3_defs = json.load(open(os.path.join(os.path.dirname(__file__),'definitions_w3.json')))
w3_code = open(code_file).readlines()

var_sector, var_subsector = get_sectors(code_file)
sector_vars = {sector: [] for sector in var_sector.values()}
for var in var_sector:
    sector_vars[var_sector[var]].append(var)
    
var_color = {v: sector_color[sector] for v, sector in var_sector.items()}



# File plot_utils.py Functions to plot variable as in world3 graphics.
import matplotlib.pyplot as plt
from matplotlib.ticker import EngFormatter
from matplotlib.image import imread

# Taken from pyworld3
def plot_world_variables(time, var_data, var_names, var_lims,
                         img_background=None,
                         title=None,
                         figsize=None,
                         dist_spines=0.09,
                         grid=False,
                         colors=None):
    """
    Plots world state from an instance of World3 or any single sector.

    """
    if not colors:
        prop_cycle = plt.rcParams['axes.prop_cycle']
        colors = prop_cycle.by_key()['color']
    else:
        colors = [colors[i] for i in var_names]
        
    var_number = len(var_data)
    
    fig, hosts = plt.subplots(1, 1, figsize=figsize)
    if not isinstance(hosts,list):
        hosts = [hosts]
    for host in hosts:
        axs = [host, ]
        for i in range(var_number-1):
            axs.append(host.twinx())

        fig.subplots_adjust(left=dist_spines*2)
        for i, ax in enumerate(axs[1:]):
            ax.spines["left"].set_position(("axes", -(i + 1)*dist_spines))
            ax.spines["left"].set_visible(True)
            ax.yaxis.set_label_position('left')
            ax.yaxis.set_ticks_position('left')

        if img_background is not None:
            im = imread(img_background)
            axs[0].imshow(im, aspect="auto",
                          extent=[time[0], time[-1],
                                  var_lims[0][0], var_lims[0][1]], cmap="gray")

        ps = []
        for ax, label, ydata, color in zip(axs, var_names, var_data, colors):
            ps.append(ax.plot(time, ydata, label=label, color=color)[0])
            axs[0].grid(grid)
            axs[0].set_xlim(time[0], time[-1])

        for ax, lim in zip(axs, var_lims):
            ax.set_ylim(lim[0], lim[1])

        for ax_ in axs:
            formatter_ = EngFormatter(places=0, sep="\N{THIN SPACE}")
            ax_.tick_params(axis='y', rotation=90)
            ax_.yaxis.set_major_locator(plt.MaxNLocator(5))
            ax_.yaxis.set_major_formatter(formatter_)

        tkw = dict(size=4, width=1.5)
        axs[0].set_xlabel("time [years]")
        axs[0].tick_params(axis='x', **tkw)
        for i, (ax, p) in enumerate(zip(axs, ps)):
            ax.set_ylabel(p.get_label(), rotation="horizontal")
            ax.yaxis.label.set_color(p.get_color())
            ax.tick_params(axis='y', colors=p.get_color(), **tkw)
            ax.yaxis.set_label_coords(-i*dist_spines, 1.01)

    if title is not None:
        fig.suptitle(title, x=0.95, ha="right", fontsize=10)

    plt.tight_layout()

# Default scales of variables
default_plot_variables_scales = {"nrfr": [0,1],
                          "iopc": [0,1e3],
                          "fpc": [0,1e3],
                          "pop": [0,16e9],
                          "ppolx": [0,32]
                         }

def plot_world_with_scales(w, variable_scales=None, title="W3 run"):
    """
    Plot world3 variables, with potentially new scales.

    Parameters
    ----------
    w: world3
        Instance to plot
    variables: dict(str, [float, float])
        Dict of Variable name and range
    title: str
        Plot title

    Returns
    -------
    """

    if not variable_scales:
        variable_scales = default_plot_variables_scales

    plot_world_variables(w.get_time(),
                    list(map(w.get_var,
                             variable_scales.keys())),
                    variable_scales.keys(),
                    variable_scales.values(),
                    figsize = (7,5),
                    grid = 1,
                    title = title)

# New scales for 03 version (p456)
colors_03 = {
    'pop':'#d62728',
    'f': '#2ca02c',
    'io': '#ff7f0e',
    'ppolx': '#9467bd',
    'nr': '#1f77b4',
    'fpc': '#2ca02c',
    'ciopc': '#ff7f0e',
    'isopc': '#bcbd22',
    'le': '#d62728',
    'hwi': '#d62728',
    'hef': '#2ca02c'
    }
scale_03_state = {
    'pop': [0, 12e9],
    'f': [0, 6e12],
    'io': [0, 4e12],
    'ppolx': [0, 40],
    'nr': [0, 2e12]
    }
scale_03_life = {
    'fpc': [0, 1000],
    'ciopc': [0, 250],
    'isopc': [0, 1000],
    'le': [0, 90],
    }
scale_03_indices = {
    'hwi': [0, 1],
    'hef': [0, 4]
    }
def plot_03_state(s, title='State'):
    plot_world_variables(s.get_time(),
                         list(map(s.get_var,
                                  scale_03_state.keys())),
                         scale_03_state.keys(),
                         scale_03_state.values(),
                         figsize = (7,5),
                         grid = 1,
                         title = title,
                         colors=colors_03)
def plot_03_life(s, title='Life level'):
    plot_world_variables(s.get_time(),
                         list(map(s.get_var,
                                  scale_03_life.keys())),
                         scale_03_life.keys(),
                         scale_03_life.values(),
                         figsize = (7,5),
                         grid = 1,
                         title = title,
                         colors=colors_03)
def plot_03_indices(s, title='Indices level'):
    plot_world_variables(s.get_time(),
                         list(map(s.get_var,
                                  scale_03_indices.keys())),
                         scale_03_indices.keys(),
                         scale_03_indices.values(),
                         figsize = (7,5),
                         grid = 1,
                         title = title,
                         colors=colors_03)
    
def plot_world_03(s, title=None, with_legend=False):
    """
    Plots world variables with 2003 way.

    """
    
    
    scales = [scale_03_state, scale_03_life, scale_03_indices]
    try:
        time = s.time
    except AttributeError as e:
        raise Exception("System must be run before plot !") from e

    
    if with_legend:
        figsize = (8,12)
    else:
        figsize = (7, 12)

    dist_spines=0.09
    grid=1
    img_background=None
    fig, hosts = plt.subplots(3, 1, figsize=figsize)
    pretitles = ['World state', 'Material life level', 'Welfare and footprint']
    for nn, (host, scale, pretitle) in enumerate(zip(hosts, scales, pretitles)):
        var_number = len(scale)
        var_names = scale.keys()
        var_data = [getattr(s,i) for i in scale]
        var_lims = scale.values()
        colors = [colors_03[i] for i in var_names]
        
        axs = [host, ]
        for i in range(var_number-1):
            axs.append(host.twinx())

        fig.subplots_adjust(left=dist_spines*2)
        for i, ax in enumerate(axs[1:]):
            ax.spines["left"].set_position(("axes", -(i + 1)*dist_spines))
            ax.spines["left"].set_visible(True)
            ax.yaxis.set_label_position('left')
            ax.yaxis.set_ticks_position('left')

        ps = []
        for ax, label, ydata, color in zip(axs, var_names, var_data, colors):
            ps.append(ax.plot(time, ydata, label=label, color=color)[0])
            axs[0].grid(grid)
            axs[0].set_xlim(time[0], time[-1])

        for ax, lim in zip(axs, var_lims):
            ax.set_ylim(lim[0], lim[1])

        for ax_ in axs:
            formatter_ = EngFormatter(places=0, sep="\N{THIN SPACE}")
            ax_.tick_params(axis='y', rotation=90)
            ax_.yaxis.set_major_locator(plt.MaxNLocator(5))
            ax_.yaxis.set_major_formatter(formatter_)
            
        tkw = dict(size=4, width=1.5)
        axs[0].set_xlabel("time [years]")
        axs[0].tick_params(axis='x', **tkw)
        for i, (ax, p) in enumerate(zip(axs, ps)):
            ax.set_ylabel(p.get_label(), rotation="horizontal")
            ax.yaxis.label.set_color(p.get_color())
            ax.tick_params(axis='y', colors=p.get_color(), **tkw)
            ax.yaxis.set_label_coords(-i*dist_spines, 1.01)
        
        axs[0].set_xticks([1900, 1950, 2000, 2050, 2100])
        axs[0].set_yticks([i*max(list(var_lims)[0])/4 for i in range(5)])
        if with_legend:
            fig.legend(ps, [s.definition(name).capitalize() for name in var_names], bbox_to_anchor = (0.97, 0.96 - nn/3), loc='upper left')
        
        if pretitle is not None:
            host.set_title(pretitle, x=0.95, ha="right", fontsize=10)
    if title is not None:
            fig.suptitle(title, fontsize=15)
    plt.tight_layout()


def plot_world_03_compare(s, s2, title=None, with_legend=False, *args, **kwargs):
    """Compare in the main variables of the worlds, plotted in the "Limits To Growth: the 30th year update" way.
    
    Parameters
    ----------
    s : 
        World3 object (with simulation already run).
    s2 :
        World3 object to compare (with simulation already run).
    
    args, kwargs :
        See pydynmao.core.plot_utils.plot_world arguments.
    """

    scales = [scale_03_state, scale_03_life, scale_03_indices]
    try:
        time = s.time
        time2 = s2.time
    except AttributeError as e:
        raise Exception("System must be run before plot !") from e

    
    if with_legend:
        figsize = (8,12)
    else:
        figsize = (7, 12)

    dist_spines=0.09
    grid=1
    img_background=None
    fig, hosts = plt.subplots(3, 1, figsize=figsize)
    pretitles = ['World state', 'Material life level', 'Welfare and footprint']
    for nn, (host, scale, pretitle) in enumerate(zip(hosts, scales, pretitles)):
        var_number = len(scale)
        var_names = scale.keys()
        var_data = [getattr(s,i) for i in scale]
        var_data_2 = [getattr(s2,i) for i in scale]
        var_lims = scale.values()
        colors = [colors_03[i] for i in var_names]
        
        axs = [host, ]
        for i in range(var_number-1):
            axs.append(host.twinx())

        fig.subplots_adjust(left=dist_spines*2)
        for i, ax in enumerate(axs[1:]):
            ax.spines["left"].set_position(("axes", -(i + 1)*dist_spines))
            ax.spines["left"].set_visible(True)
            ax.yaxis.set_label_position('left')
            ax.yaxis.set_ticks_position('left')

        ps = []
        for ax, label, ydata, ydata2, color in zip(axs, var_names, var_data, var_data_2, colors):
            ps.append(ax.plot(time, ydata, label=label, color=color)[0])
            ps.append(ax.plot(time, ydata2, label=label, color=color, linestyle='--')[0])
            axs[0].grid(grid)
            axs[0].set_xlim(time[0], time[-1])

        # Axes
        for ax, lim in zip(axs, var_lims):
            ax.set_ylim(lim[0], lim[1])

        for ax_ in axs:
            formatter_ = EngFormatter(places=0, sep="\N{THIN SPACE}")
            ax_.tick_params(axis='y', rotation=90)
            ax_.yaxis.set_major_locator(plt.MaxNLocator(5))
            ax_.yaxis.set_major_formatter(formatter_)

        # Legends & ticks
        tkw = dict(size=4, width=1.5)
        axs[0].set_xlabel("time [years]")
        axs[0].tick_params(axis='x', **tkw)
        for i, (ax, p) in enumerate(zip(axs, ps)):
            ax.set_ylabel(p.get_label(), rotation="horizontal")
            ax.yaxis.label.set_color(p.get_color())
            ax.tick_params(axis='y', colors=p.get_color(), **tkw)
            ax.yaxis.set_label_coords(-i*dist_spines, 1.01)
        
        axs[0].set_xticks([1900, 1950, 2000, 2050, 2100])
        axs[0].set_yticks([i*max(list(var_lims)[0])/4 for i in range(5)])

        
        if with_legend:
            fig.legend(ps[::2], [s.definition(name).capitalize() for name in var_names], bbox_to_anchor = (0.97, 0.96 - nn/3), loc='upper left')

        # Title
        if pretitle is not None:
            host.set_title(pretitle, x=0.95, ha="right", fontsize=10)
    if title is not None:
            fig.suptitle(title, fontsize=15)
    plt.tight_layout()



"""World2 model, originally designed by Jay Forrester in World Dynamics Hardcover – January 1, 1971.
"""
import json
import os

from pydynamo.core.system import System

w2_defs = json.load(open(os.path.join(os.path.dirname(__file__),'definitions_w2.json')))
w2_code = open(os.path.join(os.path.dirname(__file__),'code_pydynamo_w2.py')).readlines()
scales_w2={'nr': 1e12, 'ql':2,  'ci': 20e9,'p':8e9,'polr':40}


class World2(System):
    """
    A World2 object is a System object with more convenient functions and defaults, adapted for the manipulation of the Worl2 model.
    """
    def __init__(self):
        super().__init__(w2_code.copy())
        self.add_comments(w2_defs)


    def plot_world(self, **kwargs):
        """Plot world state: ressources, pollution, population, food, life quality.
        """
        self.plot(['nr', 'p', 'ql', 'fr', 'pol'], rescale=True, **kwargs)

    def run(self, N=400, dt=0.5):
        """Run with 400 steps of 1/2 year.
        """
        super().run(N, dt)



""" Functions used to get appropriate parameters in case a pydnamo "special function" (smooth, tabhl ...) is read in an equation.
"""

import ast
from astunparse import unparse
from .parse_equations import reformat_eq

# List of special functions to handle
instance_fun_names = {'smooth', 'sample', 'dlinf3', 'delay3', 'tabhl', 'step'}

   
def change_and_get_params(node, node_name):
    """Get information needed to call the special function in the `node`, and also change the `node` to include appropriate parameter names.

    Parameters
    ----------
    node : ast.Module
        Function call to handle.
    node_name : str
        Name of the node (variable or constant) which uses this function to be upated. This name is useful to determine the new name of the function (es: `tabhl_io` for the funciton `tabhl` and the updated variable `io`.
    
    Returns
    -------
    dict
        All useful information to generate the special function. Depends on the type of the function.
    """
    name = node.func.id

    # For every type of special function, different treatments apply.
    
    if name == 'tabhl':
        params = {'table': node.args[0].id,
                'val': unparse(node.args[1]),
                'x_low': unparse(node.args[2]),
                'x_high': unparse(node.args[3]),
                'x_incr': unparse(node.args[4])}
        new_fun_name = f"tabhl_{node_name}"
        params['fun'] = new_fun_name
        node.args = [node.args[1]]


    elif name in  {'smooth', 'dlinf3', 'delay3'}:
        params = {'val': node.args[0].value.id,
                  'delay': unparse(node.args[1]).strip()}
        new_fun_name = f"{name}_{node_name}"
        params['fun'] = new_fun_name
        node.args = [node.args[0], node.args[1], ast.Name('k')]

        
    elif name == 'sample':
        params = {'fun': f'sample_{node_name}',
                  'isam': unparse(node.args[2])}
        node.args.append(ast.Name('k'))

    elif name == 'step':
        params = {'fun': 'step'}
        node.args.append(ast.Name('k'))
        
        
    params['type'] = name
    return params
    
def get_dynamo_fun_params(root, node_name):
    """Get information needed to execute the equation `node`, and also change `node` with appropriate parameters.

    Parameters
    ----------
    node : ast.Module
        Equation.
    node_name : str
        Name of the node (variable or constant) which uses this function to be upated. This name is useful to determine the new name of the function (es: `tabhl_io` for the funciton `tabhl` and the updated variable `io`.
    
    Returns
    -------
    (ast.Module, dict)
        The modified node and all useful information to generate the equation.
    """
    root = root.body[0].value

    # The node is walked through to detect and change a special function call
    # Only one special DYNAMO function is allowed in an equation ...
    for node in ast.walk(root):
        if isinstance(node, ast.Call):
            if node.func.id in instance_fun_names:
                 params = change_and_get_params(node, node_name)
                 return root, params
    return (None, None)



scenarios = [
    {
        'title': 'Scenario #1: Business as usual',
        'changes': {}
    },

    {
        'title': 'Scenario #2: double -- and more accessible -- non-renewable resources',
        'changes': {
            'fcaortm': 2002,
            'nri': 2000000000000.0,
            'fcaor2t': [1.0, 0.1, 0.05, 0.05, 0.05, 0.05, 0.05, 0.05, 0.05, 0.05, 0.05],
        }
    },

    {
        'title': 'Scenario #3: Scenario #2 plus pollution technology advancing from 2002',
        'changes': {
            'pyear': 2002,
            'pptcmt': [-0.04, 0],
            'fcaortm': 2002,
            'nri': 2000000000000.0,
            'fcaor2t': [1.0, 0.1, 0.05, 0.05, 0.05, 0.05, 0.05, 0.05, 0.05, 0.05, 0.05],
        }
    },

    {
        'title': 'Scenario #4: Scenario #3 plus yield technology advancing from 2002',
        'changes': {
            'pyear': 2002,
            'lytcrmt': [0, 0.04],
            'pptcmt': [-0.04, 0],
            'fcaortm': 2002,
            'nri': 2000000000000.0,
            'fcaor2t': [1.0, 0.1, 0.05, 0.05, 0.05, 0.05, 0.05, 0.05, 0.05, 0.05, 0.05],
        }
    },

    {
        'title': 'Scenario #5: Scenario #4 plus erosion control advancing from 2002',
        'changes': {
            'fcaortm': 2002,
            'lytcrmt': [0, 0.04],
            'llmytm': 2002,
            'fcaor2t': [1.0, 0.1, 0.05, 0.05, 0.05, 0.05, 0.05, 0.05, 0.05, 0.05, 0.05],
            'pyear': 2002,
            'nri': 2000000000000.0,
            'pptcmt': [-0.04, 0]
        }
    },
    
    {
        'title': 'Scenario #6: Scenario #5 plus resource technology advancing from 2002',
        'changes': {
            'fcaortm': 2002,
            'lytcrmt': [0, 0.04],
            'llmytm': 2002,
            'fcaor2t': [1.0, 0.1, 0.05, 0.05, 0.05, 0.05, 0.05, 0.05, 0.05, 0.05, 0.05],
            'rtcmt': [0.04, 0.0],
            'pyear': 2002,
            'nri': 2000000000000.0,
            'pptcmt': [-0.04, 0],
            'pyear': 2002,
        }
    },

    {
        'title': 'Scenario #7: is Scenario #2 plus population policies from 2002',
        'changes': {
            'zpgt': 2002,
            'fcest': 2002,
            'nri': 2000000000000.0,
        }
    },

    {
        'title': 'Scenario #8: Scenario #7 plus a fixed goal for industrial output per capita from 2002, and 25 % longer life of capital',
        'changes': {
            'iopcd': 350,
            'fcest': 2002,
            'iet': 2002,
            'alsc2': 25,
            'alai2': 2.5,
            'alic2': 18,
            'pyear': 2002,
            'nri': 2000000000000.0,
            'zpgt': 2002
        }
    },

    {
        'title': 'Scenario #9: Scenario #8 \nplus resource -, pollution -, and yield – technologies\n and erosion control advancing from 2002',
        'changes': {
            'lytcrmt': [0, 0.04],
            'llmytm': 2002,
            'iopcd': 350,
            'fcest': 2002,
            'iet': 2002,
            'rtcmt': [0.04, 0.0],
            'alsc2': 25,
            'alai2': 2.5,
            'alic2': 18,
            'pyear': 2002,
            'nri': 2000000000000.0,
            'pptcmt': [-0.04, 0],
            'zpgt': 2002
        }
    },

    {
        'title': 'Scenario #10: Scenario #9 with policy implementation 20 years earlier, in 1982',
        'changes': {
            'lytcrmt': [0, 0.04],
            'llmytm': 1982,
            'iopcd': 350,
            'fcest': 2002,
            'iet': 1982,
            'rtcmt': [0.04, 0.0],
            'alsc2': 25,
            'alai2': 2.5,
            'alic2': 18,
            'pyear': 1982,
            'nri': 2000000000000.0,
            'pptcmt': [-0.04, 0],
            'zpgt': 2002
        }
    },

    {
        'title': 'Scenario #11: Scenario #9 with policy implementation 10 years later, in 2012:',
        'changes': {
            'lytcrmt': [0, 0.04],
            'llmytm': 2012,
            'iopcd': 350,
            'fcest': 2002,
            'iet': 2012,
            'rtcmt': [0.04, 0.0],
            'alsc2': 25,
            'alai2': 2.5,
            'alic2': 18,
            'pyear': 2012,
            'nri': 2000000000000.0,
            'pptcmt': [-0.04, 0],
            'zpgt': 2002
        }
    }
]

fils = [[1, 2, 3, 4, 5], [2, 7, 8, 9], [9, 10], [9, 11]]

obs = [
    "À mesure que les ressources s'épuisent, une part beaucoup plus grande des investissements industriels doit être alouée à leur extraction, et les investissements deviennent alors insuffisants, ce qui enclenche la boucle de rétroaction d'effondrement du capital industriel.",
    "La pollution générée par l'agriculture",
]
    



"""Special functions that are used in pydynamo equations.
"""
import numpy as np

def step(hght, sttm, t):
    """
    If sttm is a variable (ndarray), th step time is determined by the method
    described in section 2.4.2.
    If hght is a variable, the step function has the effect of a gate function
    that "opens" the gate allowwing step to equal hght after time sttm.
    The step will equal 0 until the step time (and any time that time is less
    than sttm).

    step = 0 if TIME < sttm
    step = hght otherwise

    Parameters
    ---------
    hght: float

    sttm: float:

    t : float
        current time value.
    """

    if sttm >= t:
        return hght
    return 0

def clip(func2, func1, t, t_switch):
    """
    Logical function used as time switch to change parameter value.

    Parameters
    ----------
    func2 : any

    func1 : any

    t : float
        current time value.
    t_switch : float
        time threshold.

    Returns
    -------
    func2 if t>t_switch, else func1.

    """
    
    if t <= t_switch:
        return func1
    else:
        return func2

class Sample:
        """
        When called, returns a the last stored value unless some intervals is reached.
        """
        def __init__(self, isam, time):
            self.current_value = isam
            self.time = time
        # Useful to let params in order to ensure they are calculated before
        def __call__(self, x_k, intval_k, isam, k):

            if k == 0:
                self.current_next_date = intval_k
                self.current_value = isam

            elif self.current_next_date <= self.time[k]:
                self.current_next_date = self.time[k] + intval_k
                self.current_value = x_k

            return self.current_value

class Interpol:
    """Custom interpolate, because scipy's one is slow and overkill"""
    def __init__(self, x_low, x_high, x_incr, table):
        self.xl = x_low
        self.xh = x_high
        self.table = table
        self.xi = x_incr

    # def __call__(self, x):
    #     if '__iter__' in dir(x):
    #         return np.array([self.call_i(xi) for xi in x])
    #     return self.call_i(x)
    def __call__(self, x):
        try:
            if x < self.xl:
                return self.table[0]
            if x >= self.xh:
                return self.table[-1]
            i = int((x - self.xl)//self.xi)
            return self.table[i] + ((x-self.xl)/self.xi - i)*(self.table[i+1] - self.table[i])
        except:
            # Case a vector is entered
            if '__iter__' in dir(x):
                return np.array([self.__call__(xi) for xi in x])



"""Functions used by a System instance to plot curves.

"""

import matplotlib.pyplot as plt
import networkx as nx
from pyvis.network import Network as pyvisNetwork

def plot(self, v_names=None, rescale=False, show_comments=True, filter_no=None, scales=None, colors=None, title='', linestyle='-', outside_legend_number=2, legend=True, **kwargs):
    """Plot the curves of the last simulation for indicated variables.

    Parameters
    ----------
    v_names : iterable(str)
        Variable to plot names. If None, all variables are plotted.

    rescale : bool
        If yes, all curves are normalized between 0 and 1.

    show_comments: bool
        If yes, comments are shown in the legend.

    filter_no: iterable(str)
        Names of variables that won't appear in the plot.

    scales: dict(str, float)
        Scales of variables. Variables are divided by their respective scales on the plot.

    colors: dict(str, str)
        Colors of each variable.

    title: str
        Title of the plot.

    linestyle: str
        Linestyle of the plot.

    outside_legend_number: int
        Number of lines from which legend is plotted outside the graph.

    legend: bool
        If yes, the legend is drawn.

    **kwargs
        Arguments for matplotlib.pyplot.plt

    Returns
    -------
    list(matplotlib.lines.Line2D)
        List of plotted lines

    """

    assert 'time' in dir(self), "No simulation have been run for the system !"

    if not v_names:
        v_names = self.get_all_variable_names()
        
    if isinstance(v_names, str):
        v_names = [v_names]
        
    if filter_no:
        v_names = [n for n in v_names if n not in filter_no]

    lines = []
    
    for name in v_names:
        v = getattr(self, name)

        if self.is_table(name):
            raise Exception(f"Error, {name} is a table and can't be plotted")
        # Case it's a constant
        try:
            v[0]
        except:
            v = v+self.get_time()*0
            
        if scales and name in scales:
            try:
                v = v/scales[name]
            except FloatingPointError:
                v = 1 + 0 * v
        elif rescale:
            v = v/max(abs(v))
        
        label = name     
        if show_comments:
            label = label + ' (' + self.definition(name).split('\n')[0].capitalize() + ')'
        try:
            color = colors[name]
        except:
            color = None
            
        lines.append(plt.plot(self.get_time(), v, label=label, color=color, linestyle=linestyle, **kwargs)[0])

    if rescale:
        plt.ylabel('Rescaled values')

    plt.xlabel('Time')
    
    if legend:
        if len(v_names) > outside_legend_number:
            plt.legend(loc='center left', bbox_to_anchor=[1, 0.8])
        else:
            plt.legend(loc='center left')
            
    plt.title(title)

    return lines
        
def show_influence_graph(self, variables=None, depth=1, show_init_val=True, in_notebook=True, options=None, colors=None):
    """Show variables influence newtork with the Pyvis library.
    
    Parameters
    ----------
    show_init_val : bool
        If True, show the initial value for a variable.
    
    in_notebook : bool
        If True, network appears as a Widget in the notebook.

    options : dict
        Pyvis options.
    
    colors : dict
        Colors of each variable and constant.
    """
    
    G = self.get_influence_graph()
    Gq = pyvisNetwork(notebook=in_notebook, directed=True)
    
    # If some variables are focused, only retrieve its neighbors
    if variables is not None:
        if isinstance(variables, str):
            variables = [variables]
        setvars = set()
        for v in variables:
            for p in nx.dfs_successors(G, v, depth+1):
                setvars.add(p)
            for p in nx.dfs_successors(G.reverse(), v, depth+1):
                setvars.add(p)

        G = G.subgraph(setvars)
    
    for a in sorted(G.nodes):
        com = self.definition(a).split('\n')[0]
        node = a
        title = f"{com}\n{self.raw_equation(a)}"
        col = colors[a] if colors and a in colors else None
        
        if show_init_val == True:
            if a in self.nodes['var']:
                title = f'{title}\nInit: {getattr(self, a)[0]:.2f}'
                    
        Gq.add_node(node, title=title, color=col)
    
    for a, b in sorted(G.edges):
        nodea, nodeb = (a, b)
        Gq.add_edge(nodea, nodeb)
    
    if options:
        Gq.show_buttons(options)
    
    return Gq


def plot_non_linearity(self, var, **kwargs):
    """Plot the non linear functions with which the variable is computed.

    Parameters
    ----------
    name : str
        Variable name.

    **kwargs
        Arguments for matplotlib.pyplot.plot

    Returns
    -------
    matplotlib.lines.Line2D
        Plotted line
    """
    
    assert 'time' in dir(self), "No simulation yet! Please run the system."
    
    x, y, ylabel, xlabel, title = self.get_tabhl_args(var)
    
    if x is not None and y is not None:
        l = plt.plot(x, y, **kwargs)[0]
        plt.ylabel(ylabel)
        plt.xlabel(xlabel)
        plt.title(title)

    return l

def plot_compare(self, s2, v_names, scales=None, rescale=False, *args, **kwargs):
    """Show the variables of 2 different systems.

        Parameters
        ----------
        s2: System
            Other system to compare whith.

        v_names: iterable(str)
            Names of variables or constant to plot.

        scales: dict(str, float)
            Scales of variables. Variables are divided by their respective scales on the plot.

        rescale: bool
            If yes, If yes, variables are normalized between 0 and 1.

        *args
            Argument list for the pydynamo.core.plot_system.plot function.

        **kwargs
            Arguments for the pydynamo.core.plot_system.plot function.
    Returns
    -------
    (list(matplotlib.lines.Line2D), list(matplotlib.lines.Line2D))
        First and second compared lines.

    """

    assert 'time' in dir(self) and 'time' in dir(s2), "No simulation yet! Please run the system before."
    
    # remove tables
    v_names = [v for v in v_names if not isinstance(getattr(self,v), list)]
    
    if not scales and rescale:
        scales= {}
        for v in v_names:
            try:
                scales[v] = max(max(getattr(self, v)), max(getattr(s2, v)))
            except TypeError:
                scales[v] = max(getattr(self, v), getattr(s2, v))

    lines = []
    
    lines.append(plot(self, v_names, *args, **kwargs, linestyle='-', scales=scales))
    plt.gca().set_prop_cycle(None)
    lines.append(plot(s2, v_names, *args, **kwargs, linestyle='--', scales=scales, legend=False))
    return lines
    
    
        



"""
Functions to convert a file in DYNAMO to pydynamo syntax.
"""

import re

def convert_dynamo_file(DYNAMO_file, pydynamo_file):
    """Create a new file with pydynamo equations from DYNAMO equations file.
    
    Parameters
    ----------
    DYNAMO_file : str
        File name of DYNAMO code.
    pydynamo_file : File name of the pydynamo code which will be created.
    """
    with open(new_file, 'w+') as nf:
        with open(filename, 'r') as f:
            for l in f.readlines():
                nf.write(dy2pydy(l))
                
def dy2pydy(l):
    """Convert a DYNAMO equation line to a pydynamo syntax equation.
    
    Parameters
    ---------
    l : str
        DYNAMO equation.
    """ 
    l = re.sub('^[ATRCLS] ', '', l)
    l = re.sub('^N (\w*)', '\\1.i', l)        
    l = re.sub('(?!<\w)smooth\((\w+?)\.k(?!\w)', 'smooth(\\1.j',l)
    l = re.sub('\.kl', '.k', l)
    l = re.sub('\.jk', '.j', l)
    return l



"""Define the System class, base class to simulate system dynamics from equations.
"""
import inspect
import numpy as np
from itertools import chain
import re

from .specials import clip, Sample, step, Interpol
from .delays import Delay3, Dlinf3, Smooth
from .parse_dynamo_functions import instance_fun_names
from .parse_system import get_system_dicts, list_from_function, list_from_file
from math import log, exp, sqrt
np.seterr(all='raise')


class System:
    """
    Base class for system dynamics.

    A System stores 3 dictionnaries that contains:
        - nodes (constants, variables, and functions)
        - equations (constant values, updating equations and initialisation equations)
        - comments about nodes.

    From this dictionnaries, it generates the updating pattern
    and run the simulation.
    """
    
    from .plot_system import plot, plot_non_linearity, plot_compare, show_influence_graph
    from .politics import new_cst_politic, new_var_politic, new_table_politic, new_politic
    from .graph import get_cst_graph, get_init_graph, get_update_graph, get_influence_graph, set_order
    from .graph import assert_cst_acyclic, assert_init_acyclic, assert_update_acyclic
    
    def __init__(self, code=None, prepare=True):
        """Initialise a System, empty or from pydynamo code.

        Parameters
        ----------
        code : str, iterable(str) or function
            If str, open the file named `code` and read pydynamo equations inside (see `parse_system.nec_from_file`). If iterable(str), each element is a pydynamo equations (see `parse_system.nec_from_lines`). If function, read pydynamo equations written inside the function (see `parse_system.nec_system_from_fun`).
        
        prepare : bool
            If True, prepare the System to run (see `System.prepare`).

        Examples
        --------
        With a file:
        Suppose there is the following lines inside the file `pydynamo_equations.py`:
        ```
        pop.k = pop.j*2 # Population
        pop.i = popi
        popi = 25 # Initial population
        ```
        >>> s = System("pydynamo_equations.py')
        
        With a list of equations:
        >>> equations_list = ['pop.k = pop.j*2', 'pop.i = popi', 'popi = 25']
        >>> s = System(equations_list)

        With a function:
        >>> def equations_function():
        >>>     pop.k = pop.j*2 # Population
        >>>     pop.i = popi
        >>>     popi = 25 # Initial population
        >>> 
        >>> s = System(equations_function)
        """
        self.comments = {}
        # If code is given, create a System with equations
        if code:
            if isinstance(code, list):
                self.code_lines = code
            elif callable(code):
                self.code_lines = list_from_function(code)
            elif isinstance(code, str):
                self.code_lines = list_from_file(code)

            self.reset_eqs(prepare)
            
        # Otherwise, create an empty System
        else:
            self.nodes = {
                'cst': set(),
                'var': set(),
                'fun': set(),
            }
            
            self.eqs = {
                'cst': dict(),
                'update': dict(),
                'init': dict()
            }
            self.code_lines = []

        self.political_changed = set()
            
    def add_equations(self, new_code_lines):
        """Add new equations to the older ones. 
        In case there is a conflict for a same variable or constant, the last equation only is remembered.
        
        Parameters
        ----------
        new_code_lines : list(str)
            Pydynamo equations to add.
        """
        self.code_lines = self.code_lines + new_code_lines
        
    def reset_eqs(self, prepare=True):
        """Set all nodes, equations and comments.
        """
        self.nodes, self.eqs, new_comments = get_system_dicts(self.code_lines)

        # Keep old comments
        for node, comment in new_comments.items():
            if comment != '':
                self.comments[node] = comment
        
        if prepare:
            self.prepare()

    def add_comments(self, comments):
        """Add comments to the System.

        Parameters
        ----------
        comments : dict(str: str)
            Each node name and its comment.
        """

        for node, comment in comments.items():
            self.comments[node] = comment
            
    # Getters    
    def get_all_variable_names(self):
        """
        Get the set of all variables.

        Returns
        -------
        set(str):
            Set of names of all variables
        """
        return set(self.nodes['var'])

    def get_var_names(self):
        """Return every variables names.
        
        Returns
        -------
        list(str)
            List of variables names.
        """
        return list(self.nodes['var'])
    
    def get_var(self, name):
        """Get the variable array.

        Parameters
        ----------
        name: str
            Name of variable.

        Returns
        -------
        np.array(float):
            Array of values of the variable for the last run.
        """

        assert name in self.nodes['var'], f"{name} is not a variable"
        return getattr(self, name)
        
    
    def get_out_nodes(self, node, with_definitions=False):
        """Returns the list of the nodes using the node to be computed.

        Parameters
        ----------
        node: str
            Name of the node
        
        with_definitions: bool
            If yes, returns a dictionnary with each node definition.
        """
        
        out_nodes = [b for (a, b) in self.get_influence_graph().out_edges(node)]
        if with_definitions:
            return {a: self.definition(a) for a in out_nodes}
        else:
            return out_nodes
        
    def get_in_nodes(self, node, with_definitions=False):
        """Returns the list of the nodes that this node needs to be computed.

        Parameters
        ----------
        node: str
            Name of the node
        
        with_definitions: bool
            If yes, returns a dictionnary with each node definition.
        """
        
        in_nodes =  [a for (a, b) in self.get_influence_graph().in_edges(node)]
        if with_definitions:
            return {a: self.definition(a) for a in in_nodes}
        else:
            return in_nodes
    
    def get_at(self, var, t):
        """Returns the value of var at time t, or an interpolation if between rwo timestep values.
        
        Parameters
        ----------
        var : str
            Variable name.
        
        t : float
            Time.
        """
        assert var in self.nodes['var'], f"{var} is not a variable"
        assert var in dir(self), "Simulation hasn't been run yet !"
        assert t >= self.initial_time and t <= self.final_time, "{t} is out of bounds {self.initial_time}, {s.final_time}"

        if t == self.final_time:
            return getattr(self, var)[-1]
        
        idx1 = np.arange(len(self.time))[self.time <= t][-1]
        dd = (self.time[idx1] - t)/self.dt
        v = getattr(self, var)
        return v[idx1]*(1 - dd) + v[idx1 + 1]*dd

    def __getitem__(self, arg):
        """Get item at a certain date. See System.get_at.
        
        Parameters
        ----------
        arg : str, float
            Variable name and time.
        """
        try:
            name, year = arg
            return self.get_at(name, year)

        except TypeError:
            return getattr(self, arg)
        
    def get_time(self):
        """Returns the time array.

        Returns
        -------
        np.array(float):
            Array of system time.
        """
        return self.time

    def get_tabhl_args(self, name):
        """
        Get indications about a tabhl function.

        Parameters
        ----------
        name: str
            Name of the variable using a tabhl function.

        Returns
        -------
        np.array, np.array, str, str, str:
            x, f(x), x label, y label, title
        """

        
        if not 'tabhl_' + name in dir(self):
            # Case it's not a tabhl function, check if it's a clip from 1 to 2
            try:
                a, b, *_ = self.eqs['update'][name]['args']['fun']['clip']['args']
                name = re.sub(r'\.[jk]', '', b.strip())
            except KeyError:
                raise Exception(f"Error, {name} is not updated with a non linear function, and is not a switch between to variables with a non linear function") from None

        aa = list(self.eqs['update'][name]['args']['var'])
        argname = aa[0][0].split('.')[0]
        f = getattr(self, 'tabhl_' + name)
        x = np.linspace(f.xl, f.xh, 100)
        ylabel = name + '\n'+ self.definition(name)
        argop = self.eqs['update'][name]['args']['fun']['tabhl']['val'].strip()
        argop = re.sub(r'^\(|\)$', '',re.sub(r'\.[jk]', '', argop)).strip()
        xlabel = argop + ('\n'+ self.definition(argname) if argop==argname else '')
        return x, f(x), ylabel, xlabel, f"{name} non-linear function"

    def definition(self, node):
        """Get the definition of a node.

        Parameters
        ----------
        node : str
            Name of the variable or constant to get the definition.
        """
        try:
            return self.comments[node]
        except:
            return ''


    def equation(self, node):
        """Returns the reformatted equation of a variable or constant.

        Parameters
        ----------
        node : str
            Name of the variable or constant to show the equation.
        """
        
        for t, eqq in self.eqs.items():
            if node in eqq:
                try:
                    if 'tabhl' in eqq[node]['raw_line']:
                        return f'{node}.k = ' + re.sub('tabhl', f'NLF_{node}t', re.sub(r'\_([jk])', '.\\1', eqq[node]['line']))
                    return self.raw_equation(node)
                except:
                    print(node, eqq)
                    return ''

    def raw_equation(self, node):
        """Returns the pydynamo raw equation of a variable or constant.

        Parameters
        ----------
        node : str
            Name of the variable or constant to get the raw pydynamo equation.
        """
        for t, eqq in self.eqs.items():
            try:
                return eqq[node]['raw_line']
            except:
                pass
        return ''

    # Infos
    def is_initialized(self, var):
        """Indicates if a var is initialized or not.

        Parameters
        ----------
        var : str
            Variable name.
        """
        return var in self.eqs['init']

    def is_table(self, name):
        """Indicates if a name is a table.
        
        Parameters
        ----------
        name : str
            Name of the node.
        """
        if name in self.nodes['cst']:
            cst = getattr(self, name)
            if isinstance(cst, np.ndarray) or isinstance(cst, list):
                return True
        return False

    # Assertions
    def assert_cst_defined(self):
        """Assert that every constant has an equation to be computed.
        """
        for c in self.nodes['cst']:
            assert c in self.eqs['cst'], f'Error: Constant {c} is not defined in any constant equation'

    def assert_update_defined(self):
        """Assert that every variable has an updating equation to be computed.
        """
        for v in self.nodes['var']:
            assert v in self.eqs['update'], f'Error: Variable {v} is not updated in any equation'

    def assert_init_defined(self):
        """Assert that every variable has an initialisation or updated equation to be computed.
        """
        for v in self.nodes['var']:
            assert (v in self.eqs['init']
            or v in self.eqs['update']) , f'Error: Variable {v} neither updated nor initialized'

    # Functions setting
    def set_fun(self, node, fun_name, args, line):
        """Set an updating, initialisation or constant setting function to the System. It evaluates a lambda function with the line equation inside.
        
        Parameters
        ----------
        node : str
            Constant or variable name.
        
        fun_name : str
            Name of function
        
        args : iterable(str)
            Arguments of the function.
        
        line : str
            Formula with wich the `node` is computed (set, initialised orupdated).
        """
        line = line.strip()
        args_str = ', '.join(args)
        fun_str = f'lambda {args_str}: {line}'
        doc = f'Get the value of {node} depending on {args}'

        # Function settings
        fun = eval(fun_str)    
        fun.__name__ = fun_name
        fun.__doc__ = doc
        fun.__line__ = line
        fun.__okdic__ = True
        fun.__custom_repr__ = f"{fun_name}({args_str})"
        fun.__str__ = lambda : f'System custom function {fun_name}({args})'

        setattr(self, fun_name, fun)

    def set_all_funs(self):
        """For each equation, set the appropriate function to the System.
        """
        for eq_type in self.eqs:
            for node in self.eqs[eq_type]:
                getattr(self, f'set_{eq_type}_fun_from_dict')(node)

    def set_cst_fun_from_dict(self, node):
        """For each constant equation, set the appropriate function to the System.
        """
        args = self.eqs['cst'][node]['args']
        line = self.eqs['cst'][node]['line']

        self.set_fun(node,
                     'set_' + node,
                     args['cst'],
                     line)

    def set_update_fun_from_dict(self, node):
        """For each updating equation, set the appropriate function to the System.
        """
        args = self.eqs['update'][node]['args']
        line = self.eqs['update'][node]['line']
        
        args_vars = ['_'.join(v_i) for v_i in args['var']]
        args_cst = list(args['cst'])
        arg_fun = []
        for fun_type in self.eqs['update'][node]['args']['fun']:
            p =  self.eqs['update'][node]['args']['fun'][fun_type]
            if p:
                arg_fun.append(p['type'])
        if arg_fun:
            arg_fun.append('k')

        self.set_fun(node,
                     'update_' + node,
                     args_vars + args_cst + arg_fun,
                     line)

        
    def set_init_fun_from_dict(self, node):
        """For each initialisation equation, set the appropriate function to the System.
        """
        args = self.eqs['init'][node]['args']
        line = self.eqs['init'][node]['line']
        args_vars = [v + '_' + i for v, i in args['var']]
        args_cst = list(args['cst'])

        for fun_type in self.eqs['init'][node]['args']['fun']:
            p =  self.eqs['init'][node]['args']['fun'][fun_type]
            if p != {}:
                arg_fun.append('k')
                arg_fun.append(p['type'])

        self.set_fun(node,
                     'init_' + node,
                     args_vars + args_cst,
                     line)

    # Simulation
    def get_cst_val(self, cst, args):
        """Get the value of a constant according to its equation and arguments.
        
        Parameters
        ----------
        cst : str
            Name of the constant.
        
        args : dict(str, value)
            Values of each arguments.
        """
        fun = getattr(self, 'set_' + cst)
        try:
            value = None
            value = fun(**args)
            if '__iter__' in dir(value):
                value = np.array(value)
            return value

        except Exception as e:
            raise(Exception(f"Error setting {cst}\n"
                            f" wih function {fun.__custom_repr__}\n"
                            f" with given args: {', '.join(args)}\n"
                            f" and line:\n"
                            f"{self.eqs['cst'][cst]['line']}\n\n"
                            f"Error: {e}\n"
                            f"Arg values: {args}\n"
                            f"Returned value: {value}")) from None

    def set_cst(self, cst, args):
        """Set a constant according to its equation and arguments.
        
        Parameters
        ----------
        cst : str
            Name of the constant.
        
        args : dict(str, value)
            Values of each arguments.
        """
        setattr(self, cst, self.get_cst_val(cst, args))

    def set_all_csts(self):
        """Set every constant constant according to its equation and arguments ONLY IF the constant is not set yet.
        """
        
        for cst in self.set_order('cst'):
            try:
                arg_names = self.eqs['cst'][cst]['args']['cst']
                args = {name: getattr(self, name) for name in arg_names}
            except KeyError as k:
                use_eqs = '\n'.join(f"{node} = {self.eqs['cst'][node]['line']}"
                                    for node in G.successors(cst))
                raise Exception(f"Error in setting constants {cst}, used in:\n"
                                f"{use_eqs}\n"
                                f"Constant {k.args[0]} is never set.") from None

            if not cst in dir(self):
                self.set_cst(cst, args)

    def generate_var(self, var, N):
        """Initialise an empty array for a variable.
        
        Parameters
        ----------
        var : str
            Variable name.
        
        N : int
            Size of simulation.
        """
        setattr(self, var, np.full(N, np.nan))

    def generate_all_vars(self, N):
        """Initialise an empty array for every variable.
        
        Parameters
        ----------
        N : int
            Size of simulation.
        """
        for v in self.nodes['var']:
            self.generate_var(v, N)


    def set_special_fun(self, p):
        """Set the special function from its parameters.
        
        p : dict
            Parameters of the function to be set.
        """
        f_type = p['type']
        for f_name, f_Class in [('delay3', Delay3),
                                ('smooth', Smooth),
                                ('dlinf3', Dlinf3)]:
            if f_type == f_name:
                fun = f_Class(self.dt)
                setattr(self, p['fun'], fun)

        if f_type == 'tabhl':
            x_low = eval(p['x_low'])
            x_high = eval(p['x_high'])
            x_incr = eval(p['x_incr'])
            table = getattr(self, p['table'])
            fun = Interpol(x_low, x_high, x_incr, table)
            setattr(self, p['fun'], fun)

        if f_type == 'sample':
            isam = eval(p['isam'])
            fun = Sample(isam, self.time)
            setattr(self, p['fun'], fun)

        if f_type == 'step' or type =='clip':
            pass

    def step(self, hght, sttm, k):
        """Step function. See specials.step.
        """
        return step(hght, sttm, self.time[k])

    def clip(self, v2, v1, t, y):
        """Clip function. See specials.clip.
        """
        return clip(v2, v1, t, y)
    
    def set_all_special_functions(self):
        """
        Generate every special dynamo functions that are stored.
        Constants (especially tables and delays) have to be initialized before.
        """
        for node in self.eqs['update']:
            fun_dic = self.eqs['update'][node]['args']['fun']
            for fun_name, fun_args in fun_dic.items():
                if fun_args != {}:
                    self.set_special_fun(fun_args)

    def _init_all(self):
        """Initialise every variable. Iterate over the initialisaition order and set the first value of each variable.
        """
        for var in self.set_order('init'):
            if self.is_initialized(var):
                update_fun = getattr(self, 'init_' + var)
                args = self.eqs['init'][var]['args']
                var_args = {v + '_' + i: getattr(self, v)[0] for v, i in args['var']}
                fun_args = dict()

            else:
                update_fun = getattr(self, 'update_' + var)
                args = self.eqs['update'][var]['args']
                var_args = {v + '_' + i: getattr(self, v)[0] for v, i in args['var']}
                fun_args = dict()

                for fun_type in self.eqs['update'][var]['args']['fun']:
                    p = self.eqs['update'][var]['args']['fun'][fun_type]
                    if p != {}:
                        fun_args[p['type']] = getattr(self, p['fun'])
                        fun_args['k'] = 0
            cst_args = {c: getattr(self, c) for c in args['cst']}

            # Assert that there is no variables considered as constants etc.
            tps = {'cst', 'var', 'fun'}
            dd = locals()
            
            for t, n, tt in ((t, n, tt) for t in tps
                         for tt in tps.difference({t})
                         for n in dd[tt + '_args']):
                # if tt == 'var':
                #         n = re.sub(r'\_[jk]', '', n)
                
                if n in self.nodes[t]:
                    line = self.eqs['update'][n]['line']
                    if n in self.eqs['init']:
                        line = self.eqs['init'][n]['line']

                    msg = ''.join((
                    f'Wrong call of {n}',
                    f' (which is of type {t})\n',
                    f'{var} = ',
                    line))
                    raise(AssertionError(msg))            
            self._update_variable(var, 0, update_fun, {**var_args, **cst_args, **fun_args}, init=True)

    def set_update_loop(self):
        """
        Set the update loop list, stored as _update_loop.
        The list contains all information we need to update each variable,
        and is ordered in the topological order of the updating graph
        """
        self._update_loop = []
        for var, i in self.set_order('update'):
            if i == 'k':
                u = {'var': var}
                try:
                    args = self.eqs['update'][var]['args']

                    u['cst_args'] = {c: getattr(self, c) for c in args['cst']}
                    u['var_args'] = {v + '_' + ni: (getattr(self, v), 0 if ni == 'k' else -1)
                                     for v, ni in args['var']}
                    u['fun_args'] = dict()
                    for fun_type in self.eqs['update'][var]['args']['fun']:
                        p = self.eqs['update'][var]['args']['fun'][fun_type]
                        u['fun_args'][p['type']] =  getattr(self, p['fun'])
                        u['fun_args']['k'] = None

                    u['update_fun'] = getattr(self, 'update_' + var)
                    self._update_loop.append(u)

                except AttributeError as e:
                    line = self.eqs['update'][var]['line']
                    raise(AttributeError(f"In updating {var}:\n"
                                         f"{var} = {line}\n"
                                         f"{e}")) from e

                # Assert that there is no variables considered as constants etc.
                tps = {'cst', 'var', 'fun'}
                for t, n, tt in ((t, n, tt) for t in tps
                         for tt in tps.difference({t})
                         for n in u[tt + '_args']):
                    if tt == 'var':
                        n = re.sub(r'\_[jk]', '', n)
                    assert n not in self.nodes[t], ''.join((
                        f'Wrong call of {n}',
                        f' (which is of type {t})\n',
                        f'{var}.{i} = ',
                        self.eqs['update'][var]['line']))


    def _update_all_fast(self, k):
        """Update every variable. Iterate over the updating order graph and set the k-th value of each variable.
        
        Parameters
        ----------
        k : int
            Current number of step.
        """

        for u in self._update_loop:
            if 'k' in u['fun_args']:
                u['fun_args']['k'] = k
            self._update_variable(u['var'], k, u['update_fun'],
                        {**{v: t[k+i] for v, (t, i) in u['var_args'].items()},
                         **u['cst_args'],
                         **u['fun_args']})

    def _update_variable(self, var, k, fun, args, init=False):
        """Update a variable at step `k` from its function and arguments.
        
        Parameters
        ----------
        var : str
            Variable name.
        
        k : int
            Step number.
        
        fun : function
            Function which updates variable.
    
        args : dict(str, value)
            Arguments of the function with their values.

        init : bool
            Indicates if it's an initialisation update.
        """
        try:
            # getattr(self, var)[k] = fun(**args)
            # For (small) gain in perfs, uncomment above and comment beside
            value = None
            array = getattr(self, var)
            value = fun(**args)
            assert value is not np.nan, "Error, value is nan"
            array[k] = value
        except Exception as e:
            if not init:
                eq = self.eqs['update'][var]['line']
                err = "Error updating"
            else:
                eq = self.eqs['init'][var]['line']
                err = "Error initialising"
            raise(Exception(f"{err} {var}\n"
                            f" at index {k}\n"
                            f" wih function {fun.__custom_repr__}\n"
                            f" with given args {', '.join(args)}\n"
                            f" and line:\n"
                            f"{eq}\n\n"
                            f"Error: {e}\n"
                            f"Arg values: {args}"
                            f"Returned value: {value}"))


    def prepare(self):
        """Assert that all equations are well defined, ant that the updating graph is acyclic. Also set all updating functions and constants.
"""
        self.assert_cst_defined()
        self.assert_init_defined()
        self.assert_update_defined()
        self.assert_cst_acyclic()
        self.set_all_funs()
        self.set_all_csts()

    def change_functions_in_dict(self):
        """
        In case some functions are re-defined by user,
        modify dependencies in parameters.
        """
        for f_type in ('update', 'init'):
            for f_name in (name for name in dir(self)
                           if name.startswith(f'{f_type}_')):
                fun = getattr(self, f_name)
                var = f_name.split('_', 1)[1]
                if ('__okdic__' not in dir(fun)) or (not fun.__okdic__):
                    self.nodes['var'].add(var)
                    args = inspect.getargs(fun.__code__).args
                    if fun.__doc__:
                        self.comments[var] = fun.__doc__
                    fun.__doc__ = f"User defined, {f_type} {var}"
                    fun.__custom_repr__ = f_name
                    dargs = {'cst':set(), 'var':set(), 'fun':dict()}
                    for arg in args:
                        if len(arg) > 2 and arg[-2:] in {'_k', '_j'}:
                           dargs['var'].add((arg[:-2], arg[-1]))
                        else:
                           dargs['cst'].add(arg)
                    self.eqs[f_type][var] = {'args':dargs,'line': None}
                    fun.__okdic__ = True

    

    def run(self, N=None, dt=1):
        """
        After preparing and before running,
        an user can redefine some constants,
        or assign update or init functions,
        with the following syntax:
        >>> system.c = 1 # New value for constant c
        >>> # New function to update v1
        >>> # using variable v1 at index j and constant c1:
        >>> def new_update(v2_j, c1):
               \"""Variable 2 documentation\"""
        >>>    return v2_j + c1
        >>> system.update_v1 = new_update # New fct for updating v1
        >>> s.run(10, 1) # Run with new parameters
        """

        # Change functions if manually modified, then check acyclicity
        self.change_functions_in_dict()
        self.assert_update_acyclic()
        self.assert_init_acyclic()


        # Set time steps, either with N steps of dt,
        # or a defined intitial and final time
        try:
            if not N:
                N = int((self.final_time - self.initial_time)/dt)
            else:
                self.final_time = self.initial_time + N * dt
        except:
            raise Exception("Either pass N or define both final_time and initial_time")
        self.time = self.initial_time + np.arange(N)*dt
        self.dt = dt
        
        self.generate_all_vars(N)
        self.set_all_special_functions()
        self.set_update_loop()
        self._init_all()
        for k in range(1, N):
            self._update_all_fast(k)

    def copy(self, prepare=True):
        """Returns a copy of the model, with 

        Parameters
        ----------
        prepare : bool
            If yes, prepare the system to run.
        """
        
        s = System(self.code_lines.copy())
        for cst in self.nodes['cst']:
            setattr(s, cst, getattr(self, cst))
        return s

    
    def get_different_csts(self):
        """Returns every variable which value is different than in the equations and their, abnd its new value.
        
        Returns 
        -------
        dict(str: (float, float))
            Dictionnary with each name, and its corresponding old and new value.
        """
        dic_cst = {cst: getattr(self, cst) for cst in self.nodes['cst'] if cst in dir(self)}
        for c in dic_cst:
            delattr(self, c)
        self.set_all_csts()
        both =  {cst: (getattr(self, cst), dic_cst[cst]) for cst in dic_cst}
        def diff(a, b):
            try:
                if a!=b:
                    return True
                return False
            except: return any(a!=b)
        return {cst: (v1, v2) for cst, (v1, v2) in both.items() if diff(v1, v2)}



# -*- coding: utf-8 -*-

__version__ = "0.1"

import pydynamo.core
from pydynamo.core import System, psdsystem

import pydynamo.world2
from pydynamo.world2 import World2

import pydynamo.world3
from pydynamo.world3 import World3

import pydynamo.utilities
from pydynamo.utilities import convert_dynamo_file



"""
Define the World3 clas

"""

from .plot_utils import plot_world_with_scales, plot_world_03, plot_world_03_compare
from .data_world3 import w3_code, w3_defs, var_color
from .scenarios_limits_to_growth import scenarios
import pydynamo.world3.data_world3

from pydynamo.core.system import System


class World3(System):
    """
    A World3 object is a System object with more convenient functions and some additional plotting functions, adapted for the manipulation of the World3 model 2003's version equations.
    """
    def __init__(self, scenario_number=2, sys=None):
        """Initialise a World3 object. By default, the scenario number is the second one, because it's the most "realistic" when we compare to the current situation (in 2022).
        """
        if not sys:
            ccode = w3_code.copy()
            changes = scenarios[scenario_number - 1]['changes']
            for cst, eq in changes.items():
                ccode.append(f'{cst} = {eq}')

        else:
            ccode = sys.code_lines.copy()
                
        super().__init__(ccode, True)
        self.add_comments(w3_defs)
        
    def copy(self):
        """Returns a copy of the system, with the same equations and constant values.
        """

        return World3(sys=self)

    def run(self, N=400, dt=0.5):
        """Run the system with 400 steps of 1/2 year.
        """
        super().run(N, dt)

    def plot_world(self, **kwargs):
        """Plot the main variables of the world, in the "Limits To Growth: the 30th years update" way. 

        Parameters
        ----------
        kwargs : 
            See pydynmao.world3.plot_utils.plot_world_03 arguments.
        """

        plot_world_03(self, with_legend=True, **kwargs)

    def plot_world_compare(self, s2, *args, **kwargs):
        """Compare in the main variables of the worlds, plotted in the "Limits To Growth: the 30th year update" way.

        Parameters
        ----------
        s2 :
            World3 object to compare (with simulation already run).

        args, kwargs :
            See pydynmao.world3.plot_utils.plot_world_03_compare arguments.
        """
        plot_world_03_compare(self, s2, with_legend=True, *args, **kwargs)
    
    def show_influence_graph(self, **kwargs):
        """See pydynamo.core.plot_system.show_influence_graph.
        """
        return super().show_influence_graph(colors=var_color, **kwargs)



"""Functions of System which handles networks.
"""
import networkx as nx
import re

# Graph
def get_cst_graph(self):
    """Get the graph of influences for constants: an arrow from constant A to constant B if B needs A to be computed.

    Returns
    -------
    networkx.DiGraph
        Graph of constant influences.
    """
    G = nx.DiGraph()
    for c in self.eqs['cst']:
        G.add_node(c)
        args = self.eqs['cst'][c]['args']
        for node_type in args:
            for node in args[node_type]:
                if node_type == 'var':
                    raise(Exception(
                        f'Constant {c} is set with a variable ({"_".join(node)}):\n'
                        f'{c} = {self.eqs["cst"][c]["line"]}'))
                if node_type == 'cst':
                    G.add_edge(node, c)
    return G

def get_init_graph(self):
    """Get the graph of influences for variables at initialisation: an arrow from variable A to variable B if B needs A to be computed.

    Returns
    -------
    networkx.DiGraph
        Graph of variable initialisation influences.
    """
    G = nx.DiGraph()
    for v in self.eqs['init']:
        G.add_node(v)
        args = self.eqs['init'][v]['args']
        for av, ai in args['var']:
            if av in self.eqs['update']:
                G.add_edge(av, v)
    for v in self.eqs['update']:
        if not self.is_initialized(v):
            G.add_node(v)
            args = self.eqs['update'][v]['args']
            for av, ai in args['var']:
                if av in self.eqs['update']:
                    G.add_edge(av, v)
    return G

def get_update_graph(self):
    """Get the graph of influences for variables and their indices at updating step: an arrow from variable (A, i) to variable (B, k) if (B, k) needs (A, i) to be computed.

    Returns
    -------
    networkx.DiGraph
        Graph of variable influences at updating step.
    """
    G = nx.DiGraph()
    for v in self.eqs['update']:
        G.add_node((v, 'k'))
        args = self.eqs['update'][v]['args']
        for av, ai in args['var']:
            if av in self.eqs['update']:
                G.add_edge((av, ai), (v, 'k'))
    return G

def get_update_graph_quotient(self):
    """Get the graph of influences for variables at updating step: an arrow from variable A to variable B if B needs A to be computed.

    Returns
    -------
    networkx.DiGraph
        Graph of variable influences at updating step.
    """
    G = nx.DiGraph()
    for v in self.eqs['update']:
        G.add_node(v)
        args = self.eqs['update'][v]['args']
        for av, _ in args['var']:
            if av in self.eqs['update']:
                G.add_edge(av, v)
    return G

def get_influence_graph(self):
    """Get the graph of influences: an arrow from A to B if B needs A (at initialisation or updating step) to be computed.

    Returns
    -------
    networkx.DiGraph
        Graph of influences.
    """
    G = nx.DiGraph()
    for v in self.eqs['update']:
        G.add_node(v)
        args = self.eqs['update'][v]['args']
        for av, _ in args['var']:
            if av in self.eqs['update']:
                G.add_edge(av, v)
        for ac in args['cst']:
            if ac in self.eqs['cst']:
                G.add_edge(ac, v)

    for v in self.eqs['init']:
        G.add_node(v)
        args = self.eqs['init'][v]['args']
        for av, _ in args['var']:
            if av in self.eqs['init']:
                G.add_edge(av, v)
        for ac in args['cst']:
            if ac in self.eqs['cst']:
                G.add_edge(ac, v)

    for c in self.eqs['cst']:
        G.add_node(c)
        args =  self.eqs['cst'][c]['args']
        for ac in args['cst']:
            if ac in self.eqs['cst']:
                G.add_edge(ac, c)
    return G

def set_order(self, gtype):
    """Returns the order to set constants, intitialize or update variables.

    Parameters
    ----------
    gtype : str
        Type of graph, either 'cst', 'init', or 'update'.
    """
    if gtype == 'cst':
        G = self.get_cst_graph()
    elif gtype == 'init':
        G = self.get_init_graph()
    elif gtype == 'update':
        G = self.get_update_graph()
    else:
        raise Exception("Wrong type of graph")
    return nx.topological_sort(G)

def assert_update_acyclic(self):
    """Assert that the updating graph is acyclic, and print the cycle in case there is some.
    """
    G = self.get_update_graph()
    b = '\\'
    assert nx.is_directed_acyclic_graph(G), \
        "Update is not acyclic:\n"\
         + "\n".join(f"{'.'.join(j)} <- {'.'.join(i)}: "
                     f"{'.'.join(j)} = {re.sub(b+'_([jk])', '.'+b+'1',self.eqs['update'][j[0]]['line'])}"
                     for i, j in reversed(nx.find_cycle(G)))\
         + "\nPlease design an update scheme that is not cyclic."

def assert_init_acyclic(self):
    """Assert that the initialisation graph is acyclic, and print the cycle in case there is some.
    """
    G = self.get_init_graph()
    if not nx.is_directed_acyclic_graph(G):
        msg = "Initialisation is not acyclic:\n"
        for i, j in reversed(nx.find_cycle(G)):
            line = self.eqs['update'][j]['line']
            if j in self.eqs['init']:
                line = self.eqs['init'][j]['line']
            msg += f"{j} <- {i}: "
            msg += f"{j}.i = {line} \n"
        msg +="Please design an initialisation scheme that is not cyclic."
        raise AssertionError(msg)

def assert_cst_acyclic(self):
    """Assert that the constant setting graph is acyclic.
    """
    assert nx.is_directed_acyclic_graph(self.get_cst_graph()),\
            ("Cycle detected for constant equations",
             nx.find_cycle(self.get_update_graph()))



from pydynamo import World3

# Commande pour importer et utiliser en direct une bibliothèque graphique
import matplotlib.pyplot as plt

# Déclaration d'une instance du premier (=1) scénario du modèle world3
w = World3(1)

# Lancement de la simulation, par défaut de 1900 à 2100
w.run()

# Affichage de l'évolution principales variables du monde, avec un titre
w.plot_world(title="Scénario 1: Business as usual")
plt.show()

w.equation('al')

# Affichage des variables al, io et nr, renormalisé (rescale) entre 0 et 1
w.plot(['al', 'io', 'nr'], rescale=True)

plt.show()

# Afficher la définition et l'équation de 'io'
print(w.definition('io'))
print(w.equation('io')) 
# Affichage des variables désirées, normalisé (rescale) entre 0 et 1, avec un titre
w.plot({'io', 'ic', 'fcaor', 'cuf', 'icor'}, rescale=True, title="Quelques variables en jeu dans le calcul du produit industriel")

plt.show()


print(w.equation('jph'))
w.plot_non_linearity('jph')

plt.show()

print(w.definition('jpht'))
print(w.jpht)


print(w.equation('fcaor'))
print(w.equation('fcaor1'))
print(w.definition('nrfr'))
w.plot_non_linearity('fcaor')
plt.show()

w_pas_de_gachis = World3(1)
print(w_pas_de_gachis.pl)
w_pas_de_gachis.pl = 0
w_pas_de_gachis.run()
w['pop', 2050]


wr2 = World3(scenario_number=1)
print(wr2.nri)

# Changement de valeur !
wr2.nri = 2*wr2.nri

wr2.run()
wr2.plot_world(title="Scénario avec 2 fois plus de ressources initiales")
plt.show()


w.show_influence_graph(variables=['pal', 'al'], depth = 1).show('exemple.html')

# Comparaison des systèmes w et w_pas_bcp_de_terres sur les variables f et al
w.plot_compare(w_pas_de_gachis, {'f', 'al'}, rescale=True)
plt.show()

# Affichage des variables sortantes de ppolx, avec leurs définitions respectives
print(w.get_out_nodes('ppolx', with_definitions=True))

# Affichage du graphique
w.show_influence_graph(variables='ppolx').show("exemple.html")
plt.show()

# Comparaison des systèmes w et wr2 sur différentes variables
w.plot_compare(wr2, w.get_out_nodes('ppolx'), rescale=True, title="Comparaison des scénarios 1 et 2 pour des variables liées à la pollution")
plt.show()


# Comparaison des systèmes w et wr2 sur différentes variables
w.plot_compare(wr2, {'ppolx', 'lmp', 'lfert'}, rescale=True, title="Comparaison des scénarios 1 et 2 pour des variables liées à la pollution")
plt.show()


# Une nouvelle instance du modèle assez burlesque
# Par défaut, cette commande prend le scénario 2, le plus 'réaliste' aujourd'hui
w_burlesque = World3()
# À partir de 1985, suite à des ateliers efficaces, on ne gâche plus de nourriture 
w_burlesque.new_politic('pl', 1985, 0)
# À partir de 2052, à cause de la 8G, la terre assimile moins bien la pollution
w_burlesque.new_politic('ahlmt', 2052, w.ahlmt/2)
# À partir de 2012, par miracle, l'indice de pollution persistente ppolx décroit de 2% par an
w_burlesque.new_politic('ppolx', 2012, 'ppolx.j - 0.02*ppolx.j*dt')
# Simulation et affichage
w_burlesque.run()
w_burlesque.plot(['pl', 'ahlm', 'ppolx'], rescale=True)
plt.show()

# Affiche l'évolution la pollution et la génération de pollution pour le scnéario 2 
wr2 = World3() # Par défaut, le scénario de World3 est le second, car c'est le plus "réaliste"
wr2.run() 
wr2.plot(['ppol', 'ppgr'], rescale=True, title='La pollution du scénario 2')
plt.show()

 # Afficher l'équation et les définitions des variables utilisées pour le calcul de la génération de pollution ppgr
print(wr2.equation('ppgr'))
print(wr2.get_in_nodes('ppgr', with_definitions=True))
# Afficher les différents facteurs de polutions, sans rescale pour pouvoir comparer
wr2.plot(['ppgr', 'ppgio', 'ppgao', 'ppgf'], title="Facteurs de pollution dans le scénario 2")
plt.show()


print(wr2.equation('ppgao'))
print(wr2.get_in_nodes('ppgao', with_definitions=True))
print(wr2.equation('aiph'))
print(wr2.get_in_nodes('aiph', with_definitions=True))

print("Valeurs des constantes à changer:", wr2.fipm, wr2.amti)
print("Équation de falm:", wr2.equation('falm'))
print("Valeur de falmt:", wr2.falmt)
wr2.plot_non_linearity('falm')
plt.show()

w_bio = World3()

# Changement de politique !
w_bio.new_politic('fipm', 2020, w.fipm/2) # Moins de matériel persistent
w_bio.new_politic('amti', 2020, w.amti/2) # Moins toxique
w_bio.new_politic('falmt', 2020, w.falmt*2) # Plus de maintenance (et donc moins d'intrants)

# Simuler et afficher
w_bio.run()
w_bio.plot_world(title="Scénario du passage au bio")
plt.show()

plt.figure(figsize=(10, 5)) # Augmente la taille de la figure pour y voir plus clair
wr2.plot_compare(w_bio, ['nr', 'iopc', 'fpc', 'le', 'ppolx'], rescale=True, title="Passage au bio: quelques différences clés")
plt.show()

print(w_bio.equation('f'))
wr2.plot_compare(w_bio, w_bio.get_in_nodes('f'), rescale=True, title="La chute de nourriture dans le passage au bio")
plt.show()

print(w.equation('al'))
w_bio.plot(['ldr', 'ler', 'lrui'], title="Les variables en jeu dans la mise en culture agricole")
plt.show()

print(w.equation('ldr'))
w_bio.plot(['tai', 'fiald', 'dcph'], rescale=True, title="Enjeux économiques agricoles")
plt.show()

print(w_bio.equation('dcph'))
print(w_bio.get_in_nodes('dcph', with_definitions=True))
w_bio.plot_non_linearity('dcph')
plt.show()

# Affiche l'ensemble des utilisation de terres
w_bio.plot(['al', 'pal', 'pali', 'uil'], title="Utilisation des terres dans le scénario bio")
plt.show()

# Copie du scénario "décroissance"

w_dec = World3()
w_dec.new_politic('io', 2002, 'io.j  - 0.05*dt*io.j')
w_dec.run()
w_dec.plot_world(title="Scénario de décroissance industrielle")

w_viable = w_dec.copy()

# Santé autonome
w_viable.new_politic('hsapct', 2002, [100, 130, 150, 165, 180, 210, 220, 235, 230])
# Agriculture sans intrants
w_viable.new_politic('lymct', 2002, [w_dec.get_at('lymc', 2000) for _ in range(len(w_dec.lymct))])
# Nature urbaine
w_viable.new_politic('cmit', 2002, w_dec.cmit*0)
# Agro-écologie
w_viable.llmytm = 2002
w_viable.new_politic('llmy2t', 2002, [1 for _ in range(len(w_dec.llmy2t))])
# Stabilisation des naissances
w_viable.zpgt = 2002
w_viable.fcest = 2002

# Simulation et affichage !
w_viable.run()
w_viable.plot_world(title="Deuxième réunion des alternatives")
plt.show()



"""Functions to add new politics in a System object.
"""
import re

def new_politic(self, name, date, new_val):
        """Implements a new politic for some constant, table or variable from a certain date. PLEASE DO NOT ADD 2 NEW POLITICS ON THE SAME VARIABLE OR CONSTANT !

        Parameters
        ----------
        name: str
            Name of a constant, table or variable we want to change.

        date: float
            date from which the new value will be activated.

        new_val: float, array or string
            If name refers to a constant, a float with the new value. If name srefers to a table, an array of the sime size as the older one. If name refers to a variable, a string with the new value.

        """
        assert name not in self.political_changed, f"New politics have already been implemented on \"{name}\". Reset the system, or apply a politic over an other variable or constant."
        if name in self.nodes['cst']:
            n = getattr(self, name)
            if '__iter__' in dir(n):
                self.new_table_politic(name[:-1], date, new_val)
            else:
                self.new_cst_politic(name, date, new_val)
            self.political_changed.add(name)
        elif name in self.nodes['var']:
            self.new_var_politic(name, date, new_val)
            self.political_changed.add(name)
        else:
            raise NameError(f'No variable or constant named "{name}" in the System')

def new_cst_politic(self, cst, date, new_value):
    """Add a new equations to `s` for the constant `cst`, which becomes a variable changing from its former value to `new_value` after the `date`.
    
    Parameters
    ----------
    s : System
        The system to implement the new politic.
    cst : str
        Constant name.
    date : float
        Date of politic application.
    new_value : float
        The new value that `cst` will take after the date `date`.
    """
    assert cst in self.eqs['cst'], f"{cst} is not a constant"
    initval= self.eqs['cst'][cst]['line']
    del self.eqs['cst'][cst]

    # New equations
    eq_clip = f"{cst}.k = clip({cst}_X2, {cst}_X1, time.k, {cst}_date) # {self.definition(cst)} " 
    eq_date = f"{cst}_date = {date} # Date of {cst} change"
    eq_cst1 = f"{cst}_X1 = {initval} # {self.definition({cst})} before {cst}_date"
    eq_cst2 = f"{cst}_X2 = {new_value} # {self.definition({cst})} after {cst}_date"

    # Change every call of cst to a call of cst.k
    for i in range(len(self.code_lines)):
        self.code_lines[i] = re.sub(r'(?<!\w)'+str(cst)+r'(?!\w)', f'{cst}.k', self.code_lines[i])
        
    self.add_equations([eq_clip, eq_date, eq_cst1, eq_cst2])
    self.reset_eqs()


def new_table_politic(self, var, date, new_value):
    """Add a new equations to `s` for the table used by the variable `var`. The table changes from its former value to `new_value` after the `date`.
    
    Parameters
    ----------
    s : System
        The system to implement the new politic.
    var : str
        Variable name.
    date : float
        Date of politic application.
    new_value : list(float)
        The new value that the table will take after the date `date`.
    """
    assert f'tabhl' in self.eqs['update'][var]['args']['fun'], f"{var} hasn't tabhl function"
    
    table_name = self.eqs['update'][var]['args']['fun']['tabhl']['table']
    table_init_val = self.eqs['cst'][table_name]['line']
    var_line = self.eqs['update'][var]['raw_line']

    # Define new equations
    eq_table_1 = f"{table_name}_X1 = {table_init_val} # {self.definition(table_name)} before {date}"
    eq_table_2 = f"{table_name}_X2 = {', '.join([str(v) for v in new_value])} # {self.definition(table_name)} after {date}"    
    eq_var_1 = var_line.replace(f'{var}.k', f'{var}_X1.k').replace(table_name, table_name + '_X1')
    eq_var_2 = var_line.replace(f'{var}.k', f'{var}_X2.k').replace(table_name, table_name + '_X2')
    eq_var = f"{var}.k = clip({var}_X2.k, {var}_X1.k, time.k, {date}) # {self.definition(var)}"

    self.add_equations([eq_table_1, eq_table_2, eq_var_1, eq_var_2, eq_var])
    self.reset_eqs()

def new_var_politic(self, var, date, eq2):
    """Add a new equations to `s` for the variable `var`, which changes from its former value to `new_value` after the `date`.
    
    Parameters
    ----------
    s : System
        The system to implement the new politic.
    var : str
        Variable name.
    date : float
        Date of politic application.
    new_value : str
        The new value, written with the pydynamo syntax, that the variable `var` will take after the date `date`.
    """
    assert var in self.eqs['update'], f"{var} is not a variable"
    first_line = self.eqs['update'][var]['raw_line'].split('=')[1]
    line_init = None
    if var in self.eqs['init']:
        line_init = self.eqs['init'][var]['raw_line'].split('=')[1]
    del self.eqs['update'][var]

    # Define new equations
    eq_clip = f"{var}.k = clip({var}_X2.k, {var}_X1.k, time.k, {var}_date) # {self.definition(var)}"
    eq_date = f"{var}_date = {date} # Date of {var} change"
    eq_var1 = f"{var}_X1.k = {first_line} # {self.definition(var)} before {var}_date"
    eq_var2 = f"{var}_X2.k = {eq2} # {self.definition(var)} after {var}_date"
    eq_init1 = ''
    eq_init2 = ''
    if line_init:
        eq_init1 = f"{var}_X1.i = {line_init}"
    eq_init2 = f"{var}_X2.i = 0"

    self.add_equations([eq_clip, eq_date, eq_var1, eq_var2, eq_init1, eq_init2])
    self.reset_eqs()

